# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_config import cfg
import oslo_messaging as messaging
from oslo_messaging.rpc import client

from qinling import context as ctx
from qinling import exceptions as exc

_TRANSPORT = None
_ENGINE_CLIENT = None


def cleanup():
    """Intended to be used by tests to recreate all RPC related objects."""

    global _TRANSPORT
    global _ENGINE_CLIENT

    _TRANSPORT = None
    _ENGINE_CLIENT = None


def get_transport():
    global _TRANSPORT

    if not _TRANSPORT:
        _TRANSPORT = messaging.get_rpc_transport(cfg.CONF)

    return _TRANSPORT


def get_engine_client():
    global _ENGINE_CLIENT

    if not _ENGINE_CLIENT:
        _ENGINE_CLIENT = EngineClient(get_transport())

    return _ENGINE_CLIENT


def _wrap_exception_and_reraise(exception):
    message = "%s: %s" % (exception.__class__.__name__, exception.args[0])

    raise exc.QinlingException(message)


def wrap_messaging_exception(method):
    """This decorator unwrap remote error in one of QinlingException.

    oslo.messaging has different behavior on raising exceptions
    when fake or rabbit transports are used. In case of rabbit
    transport it raises wrapped RemoteError which forwards directly
    to API. Wrapped RemoteError contains one of QinlingException raised
    remotely on Engine and for correct exception interpretation we
    need to unwrap and raise given exception and manually send it to
    API layer.
    """
    def decorator(*args, **kwargs):
        try:
            return method(*args, **kwargs)
        except exc.QinlingException:
            raise
        except (client.RemoteError, Exception) as e:
            if hasattr(e, 'exc_type') and hasattr(exc, e.exc_type):
                exc_cls = getattr(exc, e.exc_type)
                raise exc_cls(e.value)

            _wrap_exception_and_reraise(e)

    return decorator


class ContextSerializer(messaging.Serializer):
    def __init__(self, base):
        self._base = base

    def serialize_entity(self, context, entity):
        if not self._base:
            return entity
        return self._base.serialize_entity(context, entity)

    def deserialize_entity(self, context, entity):
        if not self._base:
            return entity
        return self._base.deserialize_entity(context, entity)

    def serialize_context(self, context):
        return context.convert_to_dict()

    def deserialize_context(self, context):
        qinling_ctx = ctx.Context.from_dict(context)
        ctx.set_ctx(qinling_ctx)

        return qinling_ctx


class EngineClient(object):
    """RPC Engine client."""

    def __init__(self, transport):
        """Constructs an RPC client for engine.

        :param transport: Messaging transport.
        """
        serializer = ContextSerializer(
            messaging.serializer.JsonPayloadSerializer())

        self.topic = cfg.CONF.engine.topic

        self._client = messaging.RPCClient(
            transport,
            messaging.Target(topic=self.topic),
            serializer=serializer
        )

    @wrap_messaging_exception
    def create_runtime(self, id):
        return self._client.prepare(topic=self.topic, server=None).cast(
            ctx.get_ctx(),
            'create_runtime',
            runtime_id=id
        )

    @wrap_messaging_exception
    def delete_runtime(self, id):
        return self._client.prepare(topic=self.topic, server=None).cast(
            ctx.get_ctx(),
            'delete_runtime',
            runtime_id=id
        )

    @wrap_messaging_exception
    def update_runtime(self, id, image=None, pre_image=None):
        return self._client.prepare(topic=self.topic, server=None).cast(
            ctx.get_ctx(),
            'update_runtime',
            runtime_id=id,
            image=image,
            pre_image=pre_image
        )

    @wrap_messaging_exception
    def get_runtime_pool(self, runtime_id):
        return self._client.prepare(topic=self.topic, server=None).call(
            ctx.get_ctx(),
            'get_runtime_pool',
            runtime_id=runtime_id
        )

    @wrap_messaging_exception
    def create_execution(self, execution_id, function_id, version, runtime_id,
                         input=None, is_sync=True):
        method_client = self._client.prepare(topic=self.topic, server=None)

        if is_sync:
            return method_client.call(
                ctx.get_ctx(),
                'create_execution',
                execution_id=execution_id,
                function_id=function_id,
                function_version=version,
                runtime_id=runtime_id,
                input=input
            )
        else:
            method_client.cast(
                ctx.get_ctx(),
                'create_execution',
                execution_id=execution_id,
                function_id=function_id,
                function_version=version,
                runtime_id=runtime_id,
                input=input
            )

    @wrap_messaging_exception
    def delete_function(self, id, version=0):
        return self._client.prepare(topic=self.topic, server=None).cast(
            ctx.get_ctx(),
            'delete_function',
            function_id=id,
            function_version=version
        )

    @wrap_messaging_exception
    def scaleup_function(self, id, runtime_id, version=0, count=1):
        return self._client.prepare(topic=self.topic, server=None).cast(
            ctx.get_ctx(),
            'scaleup_function',
            function_id=id,
            runtime_id=runtime_id,
            function_version=version,
            count=count
        )

    @wrap_messaging_exception
    def scaledown_function(self, id, version=0, count=1):
        return self._client.prepare(topic=self.topic, server=None).cast(
            ctx.get_ctx(),
            'scaledown_function',
            function_id=id,
            function_version=version,
            count=count
        )



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import os

from PIL import Image
import swiftclient
from swiftclient.exceptions import ClientException


def resize_image(image_path, resized_path):
    with Image.open(image_path) as image:
        image.thumbnail(tuple(x / 4 for x in image.size))
        image.save(resized_path)


def main(context, container_name, object_name):
    conn = swiftclient.Connection(
        session=context['os_session'],
        os_options={'region_name': 'RegionOne'},
    )

    # Download original image
    image_path = os.path.abspath('./%s' % object_name)
    _, obj_contents = conn.get_object(container_name, object_name)
    with open(image_path, 'w') as local:
        local.write(obj_contents)

    print('Downloaded object %s from container %s' %
          (object_name, container_name))

    thumb_path = os.path.abspath('./%s_resized.png' % object_name)
    resize_image(image_path, thumb_path)

    print('Resized.')

    # Create new container if needed
    new_container_name = '%s_resized' % container_name
    try:
        conn.head_container(new_container_name)
    except ClientException:
        conn.put_container(new_container_name)
        print("New container %s created." % new_container_name)

    # Upload resized image
    with open(thumb_path, 'r') as new_local:
        conn.put_object(
            new_container_name,
            object_name,
            contents=new_local,
            content_type='text/plain'
        )
    os.remove(image_path)
    os.remove(thumb_path)

    print('Uploaded object %s to container %s' %
          (object_name, new_container_name))



# Copyright 2018 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from datetime import datetime
from datetime import timedelta
import json

import mock

from qinling import context
from qinling.db import api as db_api
from qinling import status
from qinling.tests.unit.api import base
from qinling.tests.unit import base as unit_base


class TestFunctionVersionController(base.APITest):
    def setUp(self):
        super(TestFunctionVersionController, self).setUp()

        self.db_func = self.create_function()
        self.func_id = self.db_func.id

    @mock.patch('qinling.storage.file_system.FileSystemStorage.copy')
    @mock.patch('qinling.storage.file_system.FileSystemStorage.changed_since')
    @mock.patch('qinling.utils.etcd_util.get_function_version_lock')
    def test_post(self, mock_etcd_lock, mock_changed, mock_copy):
        lock = mock.Mock()
        mock_etcd_lock.return_value.__enter__.return_value = lock
        lock.is_acquired.return_value = True
        mock_changed.return_value = True

        # Getting function and versions needs to happen in a db transaction
        with db_api.transaction():
            func_db = db_api.get_function(self.func_id)
            self.assertEqual(0, len(func_db.versions))

        body = {'description': 'new version'}
        resp = self.app.post_json('/v1/functions/%s/versions' % self.func_id,
                                  body)

        self.assertEqual(201, resp.status_int)
        self._assertDictContainsSubset(resp.json, body)

        mock_changed.assert_called_once_with(unit_base.DEFAULT_PROJECT_ID,
                                             self.func_id, "fake_md5", 0)
        mock_copy.assert_called_once_with(unit_base.DEFAULT_PROJECT_ID,
                                          self.func_id, "fake_md5", 0)

        # We need to set context as it was removed after the API call
        context.set_ctx(self.ctx)

        with db_api.transaction():
            func_db = db_api.get_function(self.func_id)
            self.assertEqual(1, len(func_db.versions))

        # Verify the latest function version by calling API
        resp = self.app.get('/v1/functions/%s' % self.func_id)

        self.assertEqual(200, resp.status_int)
        self.assertEqual(1, resp.json.get('latest_version'))

    @mock.patch('qinling.storage.file_system.FileSystemStorage.changed_since')
    @mock.patch('qinling.utils.etcd_util.get_function_version_lock')
    def test_post_not_change(self, mock_etcd_lock, mock_changed):
        lock = mock.Mock()
        mock_etcd_lock.return_value.__enter__.return_value = lock
        lock.is_acquired.return_value = True
        mock_changed.return_value = False

        body = {'description': 'new version'}
        resp = self.app.post_json('/v1/functions/%s/versions' % self.func_id,
                                  body,
                                  expect_errors=True)

        self.assertEqual(403, resp.status_int)

    @mock.patch('qinling.utils.etcd_util.get_function_version_lock')
    def test_post_max_versions(self, mock_etcd_lock):
        lock = mock.Mock()
        mock_etcd_lock.return_value.__enter__.return_value = lock
        lock.is_acquired.return_value = True

        for i in range(10):
            self.create_function_version(i, function_id=self.func_id)

        resp = self.app.post_json('/v1/functions/%s/versions' % self.func_id,
                                  {},
                                  expect_errors=True)

        self.assertEqual(403, resp.status_int)

    @mock.patch('qinling.utils.etcd_util.get_function_version_lock')
    def test_post_etcd_lock_failed(self, mock_etcd_lock):
        lock = mock.Mock()
        mock_etcd_lock.return_value.__enter__.return_value = lock
        lock.is_acquired.return_value = False

        body = {'description': 'new version'}
        resp = self.app.post_json('/v1/functions/%s/versions' % self.func_id,
                                  body, expect_errors=True)

        self.assertEqual(500, resp.status_int)
        self.assertEqual("Internal server error.", resp.json['faultstring'])

    def test_get_all(self):
        db_api.increase_function_version(self.func_id, 0,
                                         description="version 1")

        resp = self.app.get('/v1/functions/%s/versions' % self.func_id)

        self.assertEqual(200, resp.status_int)
        actual = self._assert_single_item(resp.json['function_versions'],
                                          version_number=1)
        self.assertEqual("version 1", actual.get('description'))

    def test_get(self):
        db_api.increase_function_version(self.func_id, 0,
                                         description="version 1")

        resp = self.app.get('/v1/functions/%s/versions/1' % self.func_id)

        self.assertEqual(200, resp.status_int)
        self.assertEqual("version 1", resp.json.get('description'))

    @mock.patch('qinling.utils.etcd_util.delete_function')
    @mock.patch('qinling.rpc.EngineClient.delete_function')
    @mock.patch('qinling.storage.file_system.FileSystemStorage.delete')
    def test_delete(self, mock_package_delete, mock_engine_delete,
                    mock_etcd_delete):
        db_api.increase_function_version(self.func_id, 0,
                                         description="version 1")

        resp = self.app.delete('/v1/functions/%s/versions/1' % self.func_id)

        self.assertEqual(204, resp.status_int)
        mock_engine_delete.assert_called_once_with(self.func_id, version=1)
        mock_etcd_delete.assert_called_once_with(self.func_id, version=1)
        mock_package_delete.assert_called_once_with(
            unit_base.DEFAULT_PROJECT_ID,
            self.func_id, None, version=1
        )

        # We need to set context as it was removed after the API call
        context.set_ctx(self.ctx)

        with db_api.transaction():
            func_db = db_api.get_function(self.func_id)
            self.assertEqual(0, len(func_db.versions))
            self.assertEqual(0, func_db.latest_version)

    def test_delete_with_running_job(self):
        db_api.increase_function_version(self.func_id, 0,
                                         description="version 1")
        self.create_job(
            self.func_id,
            function_version=1,
            status=status.RUNNING,
            first_execution_time=datetime.utcnow(),
            next_execution_time=datetime.utcnow() + timedelta(hours=1),
        )

        resp = self.app.delete(
            '/v1/functions/%s/versions/1' % self.func_id,
            expect_errors=True
        )

        self.assertEqual(403, resp.status_int)

    def test_delete_with_webhook(self):
        db_api.increase_function_version(self.func_id, 0,
                                         description="version 1")
        self.create_webhook(self.func_id, function_version=1)

        resp = self.app.delete(
            '/v1/functions/%s/versions/1' % self.func_id,
            expect_errors=True
        )

        self.assertEqual(403, resp.status_int)

    def test_delete_with_alias(self):
        db_api.increase_function_version(self.func_id, 0,
                                         description="version 1")
        name = self.rand_name(name="alias", prefix=self.prefix)
        body = {
            'function_id': self.func_id,
            'function_version': 1,
            'name': name
        }
        db_api.create_function_alias(**body)

        resp = self.app.delete(
            '/v1/functions/%s/versions/1' % self.func_id,
            expect_errors=True
        )

        self.assertEqual(403, resp.status_int)

    @mock.patch('qinling.rpc.EngineClient.scaleup_function')
    def test_scale_up(self, scaleup_function_mock):
        db_api.increase_function_version(self.func_id, 0)

        body = {'count': 1}
        resp = self.app.post(
            '/v1/functions/%s/versions/1/scale_up' % self.func_id,
            params=json.dumps(body),
            content_type='application/json'
        )

        self.assertEqual(202, resp.status_int)
        scaleup_function_mock.assert_called_once_with(
            self.func_id,
            runtime_id=self.db_func.runtime_id,
            version=1,
            count=1
        )

    @mock.patch('qinling.utils.etcd_util.get_workers')
    @mock.patch('qinling.rpc.EngineClient.scaledown_function')
    def test_scale_down(self, scaledown_function_mock, get_workers_mock):
        db_api.increase_function_version(self.func_id, 0)
        get_workers_mock.return_value = [mock.Mock(), mock.Mock()]

        body = {'count': 1}
        resp = self.app.post(
            '/v1/functions/%s/versions/1/scale_down' % self.func_id,
            params=json.dumps(body),
            content_type='application/json'
        )

        self.assertEqual(202, resp.status_int)
        scaledown_function_mock.assert_called_once_with(self.func_id,
                                                        version=1, count=1)

    @mock.patch('qinling.utils.etcd_util.get_workers')
    @mock.patch('qinling.rpc.EngineClient.scaledown_function')
    def test_scale_down_no_need(self, scaledown_function_mock,
                                get_workers_mock):
        db_api.increase_function_version(self.func_id, 0)
        get_workers_mock.return_value = [mock.Mock()]

        body = {'count': 1}
        resp = self.app.post(
            '/v1/functions/%s/versions/1/scale_down' % self.func_id,
            params=json.dumps(body),
            content_type='application/json'
        )

        self.assertEqual(202, resp.status_int)
        scaledown_function_mock.assert_not_called()

    @mock.patch('qinling.utils.etcd_util.delete_function')
    @mock.patch('qinling.rpc.EngineClient.delete_function')
    def test_detach(self, engine_delete_function_mock,
                    etcd_delete_function_mock):
        db_api.increase_function_version(self.func_id, 0)

        resp = self.app.post(
            '/v1/functions/%s/versions/1/detach' % self.func_id
        )

        self.assertEqual(202, resp.status_int)
        engine_delete_function_mock.assert_called_once_with(self.func_id,
                                                            version=1)
        etcd_delete_function_mock.assert_called_once_with(self.func_id,
                                                          version=1)



# Copyright 2017 Catalyst IT Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json

import six
from tempest.lib.common import rest_client

urlparse = six.moves.urllib.parse


class QinlingClientBase(rest_client.RestClient):
    def __init__(self, auth_provider, **kwargs):
        super(QinlingClientBase, self).__init__(auth_provider, **kwargs)

    def get_list_objs(self, obj, params=None):
        url = '/v1/%s' % obj
        query_string = ("?%s" % urlparse.urlencode(list(params.items()))
                        if params else "")
        url += query_string

        resp, body = self.get(url)
        return resp, json.loads(body)

    def delete_obj(self, obj, id):
        return self.delete('/v1/{obj}/{id}'.format(obj=obj, id=id))

    def get_obj(self, obj, id):
        resp, body = self.get('/v1/{obj}/{id}'.format(obj=obj, id=id))

        return resp, json.loads(body)

    def post_json(self, obj, req_body, extra_headers={}):
        headers = {"Content-Type": "application/json"}
        headers = dict(headers, **extra_headers)
        url_path = '/v1/%s' % obj

        resp, body = self.post(url_path, json.dumps(req_body), headers=headers)

        return resp, json.loads(body)



# Copyright 2018 Catalyst IT Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from datetime import datetime
from datetime import timedelta

from tempest.lib import decorators

from qinling_tempest_plugin.tests import base


class JobsTest(base.BaseQinlingTest):
    name_prefix = 'JobsTest'

    def setUp(self):
        super(JobsTest, self).setUp()

        self.wait_runtime_available(self.runtime_id)
        self.function_id = self.create_function()

    @decorators.idempotent_id('68e4d562-f762-11e7-875d-00224d6b7bc1')
    def test_get_all_admin(self):
        """Admin user can get jobs of other projects"""
        job_id = self.create_job(self.function_id)

        resp, body = self.admin_client.get_resources(
            'jobs?all_projects=true'
        )
        self.assertEqual(200, resp.status)
        self.assertIn(
            job_id,
            [item['id'] for item in body['jobs']]
        )

    @decorators.idempotent_id('82a694a7-d3b5-4b6c-86e5-5fac6eae0f2a')
    def test_create_with_function_version(self):
        version = self.create_function_version(self.function_id)
        # first_execution_time is at least 1 min ahead of current time.
        first_execution_time = str(datetime.utcnow() + timedelta(seconds=90))
        job_id = self.create_job(self.function_id, version=version,
                                 first_execution_time=first_execution_time)

        # Wait for job to be finished
        self.wait_job_done(job_id)

        resp, body = self.client.get_resources(
            'executions',
            {'description': 'has:%s' % job_id}
        )
        self.assertEqual(200, resp.status)
        self.assertEqual(1, len(body['executions']))

        exec_id = body['executions'][0]['id']
        self.wait_execution_success(exec_id)

        resp, body = self.client.get_execution_log(exec_id)
        self.assertEqual(200, resp.status)
        self.assertIn('Hello, World', body)

    @decorators.idempotent_id('2ff6b90b-0432-44ec-8698-eed1c7fb7f04')
    def test_create_with_function_alias(self):
        version = self.create_function_version(self.function_id)
        function_alias = self.create_function_alias(self.function_id, version)
        # first_execution_time is at least 1 min ahead of current time.
        first_execution_time = str(datetime.utcnow() + timedelta(seconds=90))
        job_id = self.create_job(function_alias=function_alias,
                                 first_execution_time=first_execution_time)

        # Wait for job to be finished
        self.wait_job_done(job_id)

        resp, body = self.client.get_resources(
            'executions',
            {'description': 'has:%s' % job_id}
        )
        self.assertEqual(200, resp.status)
        self.assertEqual(1, len(body['executions']))

        exec_id = body['executions'][0]['id']
        self.wait_execution_success(exec_id)

        resp, body = self.client.get_execution_log(exec_id)
        self.assertEqual(200, resp.status)
        self.assertIn('Hello, World', body)



# Copyright 2017 Catalyst IT Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from datetime import datetime
from datetime import timedelta
import json

from oslo_log import log as logging
import requests
from tempest.lib import exceptions

from qinling_tempest_plugin.services import base as client_base

LOG = logging.getLogger(__name__)


class QinlingClient(client_base.QinlingClientBase):
    """Tempest REST client for Qinling."""

    def delete_resource(self, res, id, ignore_notfound=False):
        try:
            resp, _ = self.delete_obj(res, id)
            return resp
        except exceptions.NotFound:
            if ignore_notfound:
                pass
            else:
                raise

    def get_resource(self, res, id):
        resp, body = self.get_obj(res, id)

        return resp, body

    def get_resources(self, res, params=None):
        resp, body = self.get_list_objs(res, params=params)

        return resp, body

    def create_runtime(self, image, name=None, is_public=True):
        req_body = {"image": image, "is_public": is_public}

        if name:
            req_body.update({'name': name})

        resp, body = self.post_json('runtimes', req_body)

        return resp, body

    def create_function(self, code, runtime_id, name='', package_data=None,
                        entry=None, timeout=None):
        """Create function.

        Tempest rest client doesn't support multipart upload, so use requests
        lib instead. As a result, we can not use self.assertRaises function for
        negative tests.
        """
        headers = {'X-Auth-Token': self.auth_provider.get_token()}
        req_body = {
            'name': name,
            'runtime_id': runtime_id,
            'code': json.dumps(code),
            'timeout': timeout
        }
        if entry:
            req_body['entry'] = entry

        req_kwargs = {
            'headers': headers,
            'data': req_body
        }
        if package_data:
            req_kwargs.update({'files': {'package': package_data}})

        url_path = '%s/v1/functions' % (self.base_url)
        resp = requests.post(url_path, **req_kwargs)

        LOG.info('Request: %s POST %s', resp.status_code, url_path)

        return resp, json.loads(resp.text)

    def update_function(self, function_id, package_data=None, code=None,
                        entry=None, **kwargs):
        headers = {'X-Auth-Token': self.auth_provider.get_token()}

        req_body = {}
        if code:
            req_body['code'] = json.dumps(code)
        if entry:
            req_body['entry'] = entry
        req_body.update(kwargs)

        req_kwargs = {
            'headers': headers,
            'data': req_body
        }
        if package_data:
            req_kwargs.update({'files': {'package': package_data}})

        url_path = '%s/v1/functions/%s' % (self.base_url, function_id)
        resp = requests.put(url_path, **req_kwargs)

        LOG.info('Request: %s PUT %s', resp.status_code, url_path)

        return resp, json.loads(resp.text)

    def get_function(self, function_id):
        resp, body = self.get(
            '/v1/functions/{id}'.format(id=function_id),
        )

        return resp, json.loads(body)

    def download_function(self, function_id):
        return self.get('/v1/functions/%s?download=true' % function_id,
                        headers={})

    def detach_function(self, function_id, version=0):
        if version == 0:
            url = '/v1/functions/%s/detach' % function_id
        else:
            url = '/v1/functions/%s/versions/%s/detach' % \
                  (function_id, version)

        return self.post(url, None, headers={})

    def create_execution(self, function_id=None, alias_name=None, input=None,
                         sync=True, version=0):
        """Create execution.

        alias_name takes precedence over function_id.
        """
        if alias_name:
            req_body = {
                'function_alias': alias_name,
                'sync': sync,
                'input': input
            }
        elif function_id:
            req_body = {
                'function_id': function_id,
                'function_version': version,
                'sync': sync,
                'input': input
            }
        else:
            raise Exception("Either alias_name or function_id must be "
                            "provided.")

        resp, body = self.post_json('executions', req_body)

        return resp, body

    def get_execution_log(self, execution_id):
        return self.get('/v1/executions/%s/log' % execution_id,
                        headers={'Accept': 'text/plain'})

    def get_function_workers(self, function_id, version=0):
        q_params = None
        if version > 0:
            q_params = "/?function_version=%s" % version

        url = 'functions/%s/workers' % function_id
        if q_params:
            url += q_params

        return self.get_resources(url)

    def create_webhook(self, function_id=None, function_alias=None,
                       version=0):
        """Create webhook.

        function_alias takes precedence over function_id.
        """
        if function_alias:
            req_body = {'function_alias': function_alias}
        elif function_id:
            req_body = {
                'function_id': function_id,
                'function_version': version
            }
        else:
            raise Exception("Either function_alias or function_id must be "
                            "provided.")
        resp, body = self.post_json('webhooks', req_body)
        return resp, body

    def create_job(self, function_id=None, function_alias=None, version=0,
                   first_execution_time=None):
        """Create job.

        function_alias takes precedence over function_id.
        """
        if function_alias:
            req_body = {'function_alias': function_alias}
        elif function_id:
            req_body = {
                'function_id': function_id,
                'function_version': version
            }
        else:
            raise Exception("Either function_alias or function_id must be "
                            "provided.")

        if not first_execution_time:
            first_execution_time = str(
                datetime.utcnow() + timedelta(hours=1)
            )
        req_body.update({'first_execution_time': first_execution_time})

        resp, body = self.post_json('jobs', req_body)
        return resp, body

    def create_function_version(self, function_id, description=None):
        req_body = {}
        if description is not None:
            req_body['description'] = description

        resp, body = self.post_json(
            'functions/%s/versions' % function_id,
            req_body
        )

        return resp, body

    def delete_function_version(self, function_id, version,
                                ignore_notfound=False):
        try:
            resp, _ = self.delete(
                '/v1/functions/{id}/versions/{version}'.format(
                    id=function_id, version=version)
            )
            return resp
        except exceptions.NotFound:
            if ignore_notfound:
                pass
            else:
                raise

    def get_function_version(self, function_id, version):
        resp, body = self.get(
            '/v1/functions/%s/versions/%s' % (function_id, version),
        )

        return resp, json.loads(body)

    def get_function_versions(self, function_id):
        resp, body = self.get(
            '/v1/functions/%s/versions' % (function_id),
        )

        return resp, json.loads(body)

    def create_function_alias(self, name, function_id,
                              function_version=0, description=None):
        req_body = {
            'function_id': function_id,
            'function_version': function_version,
            'name': name
        }
        if description is not None:
            req_body['description'] = description

        resp, body = self.post_json('/aliases', req_body)

        return resp, body

    def delete_function_alias(self, alias_name, ignore_notfound=False):
        try:
            resp, _ = self.delete('/v1/aliases/%s' % alias_name)
            return resp
        except exceptions.NotFound:
            if ignore_notfound:
                pass
            else:
                raise

    def get_function_alias(self, alias_name):
        resp, body = self.get('/v1/aliases/%s' % alias_name)

        return resp, json.loads(body)

    def update_function_alias(self, alias_name, function_id=None,
                              function_version=None, description=None):
        req_body = {}
        if function_id is not None:
            req_body['function_id'] = function_id
        if function_version is not None:
            req_body['function_version'] = function_version
        if description is not None:
            req_body['description'] = description

        resp, body = self.put_json('/v1/aliases/%s' % alias_name, req_body)

        return resp, body



# Copyright 2018 AWCloud Software Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import mock
import os

from oslo_config import cfg

from qinling import config
from qinling import exceptions as exc
from qinling.storage import file_system
from qinling.tests.unit import base
from qinling.utils import common

CONF = cfg.CONF
FAKE_STORAGE_PATH = 'TMP_DIR'


class TestFileSystemStorage(base.BaseTest):
    def setUp(self):
        super(TestFileSystemStorage, self).setUp()
        CONF.register_opts(config.storage_opts, config.STORAGE_GROUP)
        self.override_config('file_system_dir', FAKE_STORAGE_PATH, 'storage')
        self.project_id = base.DEFAULT_PROJECT_ID
        self.storage = file_system.FileSystemStorage(CONF)

    @mock.patch('oslo_utils.fileutils.ensure_tree')
    @mock.patch('os.rename')
    @mock.patch('qinling.storage.file_system.open')
    @mock.patch('zipfile.is_zipfile')
    def test_store(self, is_zipfile_mock, open_mock, rename_mock,
                   ensure_tree_mock):
        is_zipfile_mock.return_value = True
        fake_fd = mock.Mock()
        open_mock.return_value.__enter__.return_value = fake_fd
        function = self.rand_name('function', prefix='TestFileSystemStorage')
        # For python3, data should be encoded into bytes before hashing.
        function_data = "Some data".encode('utf8')
        md5 = common.md5(content=function_data)

        package_updated, ret_md5 = self.storage.store(
            self.project_id, function, function_data
        )

        self.assertTrue(package_updated)
        self.assertEqual(md5, ret_md5)

        temp_package_path = os.path.join(FAKE_STORAGE_PATH, self.project_id,
                                         '%s.zip.new' % function)
        package_path = os.path.join(
            FAKE_STORAGE_PATH,
            file_system.PACKAGE_PATH_TEMPLATE % (self.project_id, function,
                                                 md5)
        )
        ensure_tree_mock.assert_called_once_with(
            os.path.join(FAKE_STORAGE_PATH, self.project_id)
        )
        fake_fd.write.assert_called_once_with(function_data)
        is_zipfile_mock.assert_called_once_with(temp_package_path)
        rename_mock.assert_called_once_with(temp_package_path, package_path)

    @mock.patch('oslo_utils.fileutils.ensure_tree')
    @mock.patch('os.path.exists')
    def test_store_zip_exists(self, exists_mock, ensure_tree_mock):
        function = self.rand_name('function', prefix='TestFileSystemStorage')
        function_data = "Some data".encode('utf8')
        md5 = common.md5(content=function_data)
        exists_mock.return_value = True

        package_updated, ret_md5 = self.storage.store(
            self.project_id, function, function_data
        )

        self.assertFalse(package_updated)
        self.assertEqual(md5, ret_md5)

        package_path = os.path.join(
            FAKE_STORAGE_PATH,
            file_system.PACKAGE_PATH_TEMPLATE % (self.project_id, function,
                                                 md5)
        )

        exists_mock.assert_called_once_with(package_path)

    @mock.patch('oslo_utils.fileutils.ensure_tree')
    def test_store_md5_mismatch(self, ensure_tree_mock):
        function = self.rand_name('function', prefix='TestFileSystemStorage')
        # For python3, data should be encoded into bytes before hashing.
        function_data = "Some data".encode('utf8')
        not_a_md5sum = "Not a md5sum"

        self.assertRaisesRegex(
            exc.InputException,
            "^Package md5 mismatch\.$",
            self.storage.store,
            self.project_id, function, function_data, md5sum=not_a_md5sum)

        ensure_tree_mock.assert_called_once_with(
            os.path.join(FAKE_STORAGE_PATH, self.project_id))

    @mock.patch('oslo_utils.fileutils.delete_if_exists')
    @mock.patch('oslo_utils.fileutils.ensure_tree')
    @mock.patch('qinling.storage.file_system.open')
    @mock.patch('zipfile.is_zipfile')
    def test_store_invalid_zip_package(
        self, is_zipfile_mock, open_mock,
        ensure_tree_mock, delete_if_exists_mock
    ):
        is_zipfile_mock.return_value = False
        fake_fd = mock.Mock()
        open_mock.return_value.__enter__.return_value = fake_fd
        function = self.rand_name('function', prefix='TestFileSystemStorage')
        # For python3, data should be encoded into bytes before hashing.
        function_data = "Some data".encode('utf8')

        self.assertRaisesRegex(
            exc.InputException,
            "^Package is not a valid ZIP package\.$",
            self.storage.store,
            self.project_id, function, function_data)

        ensure_tree_mock.assert_called_once_with(
            os.path.join(FAKE_STORAGE_PATH, self.project_id))
        fake_fd.write.assert_called_once_with(function_data)
        delete_if_exists_mock.assert_called_once_with(
            os.path.join(FAKE_STORAGE_PATH, self.project_id,
                         '%s.zip.new' % function))

    @mock.patch('os.path.exists')
    @mock.patch('qinling.storage.file_system.open')
    def test_retrieve(self, open_mock, exists_mock):
        exists_mock.return_value = True
        fake_fd = mock.Mock()
        open_mock.return_value = fake_fd
        function = self.rand_name('function', prefix='TestFileSystemStorage')

        ret = self.storage.retrieve(self.project_id, function, "fake_md5")

        package_path = os.path.join(
            FAKE_STORAGE_PATH,
            file_system.PACKAGE_PATH_TEMPLATE % (self.project_id, function,
                                                 "fake_md5")
        )
        exists_mock.assert_called_once_with(package_path)
        open_mock.assert_called_once_with(package_path, 'rb')
        self.assertEqual(fake_fd, ret)

    @mock.patch('os.path.exists')
    def test_retrieve_package_not_found(self, exists_mock):
        exists_mock.return_value = False
        function = self.rand_name('function', prefix='TestFileSystemStorage')

        self.assertRaisesRegex(
            exc.StorageNotFoundException,
            "^Package of function %s for project %s not found\.$" % (
                function, self.project_id),
            self.storage.retrieve,
            self.project_id,
            function,
            "fake_md5"
        )

        package_path = os.path.join(
            FAKE_STORAGE_PATH,
            file_system.PACKAGE_PATH_TEMPLATE % (self.project_id, function,
                                                 "fake_md5")
        )
        exists_mock.assert_called_once_with(package_path)

    @mock.patch('qinling.storage.file_system.open')
    @mock.patch('os.path.exists')
    @mock.patch('os.listdir')
    def test_retrieve_version(self, mock_list, mock_exist, mock_open):
        function = "fake_function_id"
        version = 1
        md5 = "md5"
        mock_list.return_value = ["%s_%s_%s.zip" % (function, version, md5)]
        mock_exist.return_value = True

        self.storage.retrieve(self.project_id, function, None,
                              version=version)

        version_zip = os.path.join(FAKE_STORAGE_PATH, self.project_id,
                                   "%s_%s_%s.zip" % (function, version, md5))

        mock_exist.assert_called_once_with(version_zip)

    @mock.patch('os.listdir')
    def test_retrieve_version_not_found(self, mock_list):
        function = "fake_function_id"
        version = 1
        mock_list.return_value = [""]

        self.assertRaises(
            exc.StorageNotFoundException,
            self.storage.retrieve,
            function,
            self.project_id,
            None,
            version=version
        )

    @mock.patch('os.path.exists')
    @mock.patch('os.remove')
    def test_delete(self, remove_mock, exists_mock):
        exists_mock.return_value = True
        function = self.rand_name('function', prefix='TestFileSystemStorage')

        self.storage.delete(self.project_id, function, "fake_md5")

        package_path = os.path.join(
            FAKE_STORAGE_PATH,
            file_system.PACKAGE_PATH_TEMPLATE % (self.project_id, function,
                                                 "fake_md5")
        )
        exists_mock.assert_called_once_with(package_path)
        remove_mock.assert_called_once_with(package_path)

    @mock.patch('os.path.exists')
    @mock.patch('os.remove')
    @mock.patch('os.listdir')
    def test_delete_with_version(self, mock_list, remove_mock, exists_mock):
        exists_mock.return_value = True
        function = self.rand_name('function', prefix='TestFileSystemStorage')
        version = 1
        mock_list.return_value = ["%s_%s_md5.zip" % (function, version)]

        self.storage.delete(self.project_id, function, "fake_md5", version=1)

        package_path = os.path.join(
            FAKE_STORAGE_PATH,
            self.project_id,
            file_system.PACKAGE_VERSION_TEMPLATE % (function, version, "md5")
        )
        exists_mock.assert_called_once_with(package_path)
        remove_mock.assert_called_once_with(package_path)

    @mock.patch('os.path.exists')
    @mock.patch('os.remove')
    def test_delete_package_not_exists(self, remove_mock, exists_mock):
        exists_mock.return_value = False
        function = self.rand_name('function', prefix='TestFileSystemStorage')

        self.storage.delete(self.project_id, function, "fake_md5")

        package_path = os.path.join(
            FAKE_STORAGE_PATH,
            file_system.PACKAGE_PATH_TEMPLATE % (self.project_id, function,
                                                 "fake_md5")
        )
        exists_mock.assert_called_once_with(package_path)
        remove_mock.assert_not_called()

    def test_changed_since_first_version(self):
        ret = self.storage.changed_since(self.project_id, "fake_function",
                                         "fake_md5", 0)

        self.assertTrue(ret)

    @mock.patch('os.path.exists')
    def test_changed_since_exists(self, mock_exists):
        mock_exists.return_value = True

        ret = self.storage.changed_since(self.project_id, "fake_function",
                                         "fake_md5", 1)

        self.assertFalse(ret)

        expect_path = os.path.join(FAKE_STORAGE_PATH, self.project_id,
                                   "fake_function_1_fake_md5.zip")

        mock_exists.assert_called_once_with(expect_path)

    @mock.patch('os.path.exists')
    def test_changed_since_not_exists(self, mock_exists):
        mock_exists.return_value = False

        ret = self.storage.changed_since(self.project_id, "fake_function",
                                         "fake_md5", 1)

        self.assertTrue(ret)

        expect_path = os.path.join(FAKE_STORAGE_PATH, self.project_id,
                                   "fake_function_1_fake_md5.zip")

        mock_exists.assert_called_once_with(expect_path)

    @mock.patch("shutil.copyfile")
    def test_copy(self, mock_copy):
        self.storage.copy(self.project_id, "fake_function", "fake_md5", 0)

        expect_src = os.path.join(FAKE_STORAGE_PATH, self.project_id,
                                  "fake_function_fake_md5.zip")
        expect_dest = os.path.join(FAKE_STORAGE_PATH, self.project_id,
                                   "fake_function_1_fake_md5.zip")

        mock_copy.assert_called_once_with(expect_src, expect_dest)



# Copyright 2018 OpenStack Foundation.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""add cpu column for functions

Revision ID: 003
Revises: 002
Create Date: 2018-04-19 06:03:46.687706

"""

# revision identifiers, used by Alembic.
revision = '003'
down_revision = '002'

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        'functions',
        sa.Column('cpu', sa.Integer(), nullable=False)
    )

    op.alter_column(
        'functions',
        'memory_size',
        existing_type=sa.Integer(),
        nullable=False
    )
    # ### end Alembic commands ###



# Copyright 2017 Catalyst IT Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators

from qinling_tempest_plugin.tests import base


class RuntimesTest(base.BaseQinlingTest):
    name_prefix = 'RuntimesTest'
    create_runtime = False

    @decorators.idempotent_id('fdc2f07f-dd1d-4981-86d3-5bc7908d9a9b')
    def test_crud_runtime(self):
        name = data_utils.rand_name('runtime', prefix=self.name_prefix)
        resp, body = self.admin_client.create_runtime(self.image, name)

        self.assertEqual(201, resp.status)
        self.assertEqual(name, body['name'])

        runtime_id = body['id']
        self.addCleanup(self.admin_client.delete_resource, 'runtimes',
                        runtime_id, ignore_notfound=True)

        # Get runtimes
        resp, body = self.client.get_resources('runtimes')

        self.assertEqual(200, resp.status)
        self.assertIn(
            runtime_id,
            [runtime['id'] for runtime in body['runtimes']]
        )

        # Wait for runtime to be available
        # We don't have to check k8s resource, if runtime's status has changed
        # to available, then kubernetes deployment is assumed to be ok.
        self.wait_runtime_available(runtime_id)

        # Delete runtime
        resp = self.admin_client.delete_resource('runtimes', runtime_id)

        self.assertEqual(204, resp.status)

    @decorators.idempotent_id('c1db56bd-c3a8-4ca6-9482-c362fd492db0')
    def test_create_private_runtime(self):
        """Private runtime test.

        Admin user creates a private runtime which can not be used by other
        projects.
        """
        name = data_utils.rand_name('runtime', prefix=self.name_prefix)
        resp, body = self.admin_client.create_runtime(
            self.image, name, is_public=False
        )

        self.assertEqual(201, resp.status)
        self.assertEqual(name, body['name'])
        self.assertFalse(body['is_public'])

        runtime_id = body['id']
        self.addCleanup(self.admin_client.delete_resource, 'runtimes',
                        runtime_id, ignore_notfound=True)

        # Get runtimes
        resp, body = self.client.get_resources('runtimes')

        self.assertEqual(200, resp.status)
        self.assertNotIn(
            runtime_id,
            [runtime['id'] for runtime in body['runtimes']]
        )



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import contextlib
import sys
import threading

from oslo_config import cfg
from oslo_db import exception as oslo_db_exc
from oslo_db.sqlalchemy import utils as db_utils
from oslo_log import log as logging
import sqlalchemy as sa

from qinling import context
from qinling.db import base as db_base
from qinling.db.sqlalchemy import filters as db_filters
from qinling.db.sqlalchemy import model_base
from qinling.db.sqlalchemy import models
from qinling import exceptions as exc
from qinling import status

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

_SCHEMA_LOCK = threading.RLock()
_initialized = False


def get_backend():
    """Consumed by openstack common code.

    The backend is this module itself.
    :return: Name of db backend.
    """
    return sys.modules[__name__]


def setup_db():
    global _initialized

    with _SCHEMA_LOCK:
        if _initialized:
            return

        try:
            models.Function.metadata.create_all(db_base.get_engine())

            _initialized = True
        except sa.exc.OperationalError as e:
            raise exc.DBError("Failed to setup database: %s" % str(e))


def drop_db():
    global _initialized

    with _SCHEMA_LOCK:
        if not _initialized:
            return

        try:
            models.Function.metadata.drop_all(db_base.get_engine())

            _initialized = False
        except Exception as e:
            raise exc.DBError("Failed to drop database: %s" % str(e))


def start_tx():
    db_base.start_tx()


def commit_tx():
    db_base.commit_tx()


def rollback_tx():
    db_base.rollback_tx()


def end_tx():
    db_base.end_tx()


@contextlib.contextmanager
def transaction():
    start_tx()

    try:
        yield
        commit_tx()
    finally:
        end_tx()


def _secure_query(model, *columns):
    query = db_base.model_query(model, columns)

    if not issubclass(model, model_base.QinlingSecureModelBase):
        return query

    if model == models.Runtime:
        query_criterion = sa.or_(
            model.project_id == context.get_ctx().projectid,
            model.is_public
        )
    else:
        query_criterion = model.project_id == context.get_ctx().projectid

    query = query.filter(query_criterion)

    return query


def _paginate_query(model, limit=None, marker=None, sort_keys=None,
                    sort_dirs=None, query=None):
    if not query:
        query = _secure_query(model)

    sort_keys = sort_keys if sort_keys else []

    if 'id' not in sort_keys:
        sort_keys.append('id')
        sort_dirs.append('asc') if sort_dirs else None

    query = db_utils.paginate_query(
        query,
        model,
        limit,
        sort_keys,
        marker=marker,
        sort_dirs=sort_dirs
    )

    return query


def _get_collection(model, insecure=False, limit=None, marker=None,
                    sort_keys=None, sort_dirs=None, fields=None, **filters):
    columns = (
        tuple([getattr(model, f) for f in fields if hasattr(model, f)])
        if fields else ()
    )

    query = (db_base.model_query(model, columns) if insecure
             else _secure_query(model, *columns))
    query = db_filters.apply_filters(query, model, **filters)
    query = _paginate_query(
        model,
        limit,
        marker,
        sort_keys,
        sort_dirs,
        query
    )

    try:
        return query.all()
    except Exception as e:
        raise exc.DBError(
            "Failed when querying database, error type: %s, "
            "error message: %s" % (e.__class__.__name__, str(e))
        )


def _get_collection_sorted_by_time(model, insecure=False, fields=None,
                                   sort_keys=['created_at'], **kwargs):
    return _get_collection(
        model=model,
        insecure=insecure,
        sort_keys=sort_keys,
        fields=fields,
        **kwargs
    )


@db_base.insecure_aware()
def _get_db_object_by_id(model, id, insecure=None):
    query = db_base.model_query(model) if insecure else _secure_query(model)

    return query.filter_by(id=id).first()


@db_base.insecure_aware()
def _delete_all(model, insecure=None, **kwargs):
    # NOTE(kong): If we use 'in_' operator in _secure_query(), delete()
    # method will raise error with default parameter. Please refer to
    # http://docs.sqlalchemy.org/en/rel_1_0/orm/query.html#sqlalchemy.orm.query.Query.delete
    query = db_base.model_query(model) if insecure else _secure_query(model)
    query.filter_by(**kwargs).delete(synchronize_session="fetch")


@db_base.insecure_aware()
def _get_db_object_by_name(model, name, insecure=None):
    query = db_base.model_query(model) if insecure else _secure_query(model)

    return query.filter_by(name=name).first()


@db_base.session_aware()
def conditional_update(model, values, expected_values, insecure=False,
                       filters=None, session=None):
    """Compare-and-swap conditional update SQLAlchemy implementation."""
    filters = filters or {}
    filters.update(expected_values)
    query = (db_base.model_query(model) if insecure else _secure_query(model))
    query = db_filters.apply_filters(query, model, **filters)
    update_args = {'synchronize_session': False}

    # Return True if we were able to change any DB entry, False otherwise
    result = query.update(values, **update_args)

    return 0 != result


@db_base.insecure_aware()
@db_base.session_aware()
def get_function(id, insecure=None, session=None):
    function = _get_db_object_by_id(models.Function, id, insecure=insecure)

    if not function:
        raise exc.DBEntityNotFoundError("Function not found [id=%s]" % id)

    return function


@db_base.session_aware()
def get_functions(session=None, **kwargs):
    return _get_collection_sorted_by_time(models.Function, **kwargs)


@db_base.session_aware()
def create_function(values, session=None):
    func = models.Function()
    func.update(values.copy())

    try:
        func.save(session=session)
    except oslo_db_exc.DBDuplicateEntry as e:
        raise exc.DBError(
            "Duplicate entry for Function: %s" % e.columns
        )

    return func


@db_base.session_aware()
def update_function(id, values, session=None):
    function = get_function(id)
    function.update(values.copy())

    return function


@db_base.session_aware()
def delete_function(id, session=None):
    function = get_function(id)

    session.delete(function)


@db_base.session_aware()
def delete_functions(session=None, **kwargs):
    return _delete_all(models.Function, **kwargs)


@db_base.session_aware()
def create_runtime(values, session=None):
    runtime = models.Runtime()
    runtime.update(values.copy())

    try:
        runtime.save(session=session)
    except oslo_db_exc.DBDuplicateEntry as e:
        raise exc.DBError(
            "Duplicate entry for Runtime: %s" % e.columns
        )

    return runtime


@db_base.session_aware()
def get_runtime(id, session=None):
    model = models.Runtime
    filters = sa.and_(
        model.id == id,
        sa.or_(model.project_id == context.get_ctx().projectid,
               model.is_public),
    )
    runtime = db_base.model_query(model).filter(filters).first()

    if not runtime:
        raise exc.DBEntityNotFoundError("Runtime not found [id=%s]" % id)

    return runtime


@db_base.session_aware()
def get_runtimes(session=None, **kwargs):
    return _get_collection_sorted_by_time(models.Runtime, **kwargs)


@db_base.session_aware()
def delete_runtime(id, session=None):
    # Because we don't allow normal user to delete runtime in api layer, so it
    # is safe to get runtime here
    runtime = get_runtime(id)
    session.delete(runtime)


@db_base.session_aware()
def update_runtime(id, values, session=None):
    # Because we don't allow normal user to update runtime in api layer, so it
    # is safe to get runtime here
    runtime = get_runtime(id)
    runtime.update(values.copy())

    return runtime


@db_base.insecure_aware()
@db_base.session_aware()
def delete_runtimes(session=None, insecure=None, **kwargs):
    return _delete_all(models.Runtime, insecure=insecure, **kwargs)


@db_base.session_aware()
def create_execution(values, session=None):
    execution = models.Execution()
    execution.update(values.copy())

    try:
        execution.save(session=session)
    except oslo_db_exc.DBDuplicateEntry as e:
        raise exc.DBError(
            "Duplicate entry for Execution: %s" % e.columns
        )

    return execution


@db_base.insecure_aware()
@db_base.session_aware()
def get_execution(id, insecure=None, session=None):
    execution = _get_db_object_by_id(models.Execution, id, insecure=insecure)

    if not execution:
        raise exc.DBEntityNotFoundError("Execution not found [id=%s]" % id)

    return execution


@db_base.session_aware()
def update_execution(id, values, session=None):
    execution = get_execution(id)
    execution.update(values.copy())

    return execution


@db_base.session_aware()
def get_executions(session=None, **kwargs):
    return _get_collection_sorted_by_time(models.Execution, **kwargs)


@db_base.session_aware()
def delete_execution(id, session=None):
    execution = get_execution(id)

    session.delete(execution)


@db_base.insecure_aware()
@db_base.session_aware()
def delete_executions(session=None, insecure=None, **kwargs):
    return _delete_all(models.Execution, insecure=insecure, **kwargs)


@db_base.session_aware()
def create_job(values, session=None):
    job = models.Job()
    job.update(values)

    try:
        job.save(session=session)
    except oslo_db_exc.DBDuplicateEntry as e:
        raise exc.DBError(
            "Duplicate entry for Job: %s" % e.columns
        )

    return job


@db_base.session_aware()
def get_job(id, session=None):
    job = _get_db_object_by_id(models.Job, id)
    if not job:
        raise exc.DBEntityNotFoundError("Job not found [id=%s]" % id)

    return job


@db_base.session_aware()
def delete_job(id, session=None):
    get_job(id)

    # Delete the job by ID and get the affected row count.
    table = models.Job.__table__
    result = session.execute(table.delete().where(table.c.id == id))

    return result.rowcount


@db_base.session_aware()
def update_job(id, values, session=None):
    job = get_job(id)
    job.update(values.copy())

    return job


@db_base.session_aware()
def get_next_jobs(before, session=None):
    return _get_collection(
        models.Job, insecure=True, sort_keys=['next_execution_time'],
        sort_dirs=['asc'], next_execution_time={'lt': before},
        status=status.RUNNING
    )


@db_base.session_aware()
def get_jobs(session=None, **kwargs):
    return _get_collection_sorted_by_time(models.Job, **kwargs)


@db_base.insecure_aware()
@db_base.session_aware()
def delete_jobs(session=None, insecure=None, **kwargs):
    return _delete_all(models.Job, insecure=insecure, **kwargs)


@db_base.session_aware()
def create_webhook(values, session=None):
    webhook = models.Webhook()
    webhook.update(values.copy())

    try:
        webhook.save(session=session)
    except oslo_db_exc.DBDuplicateEntry as e:
        raise exc.DBError(
            "Duplicate entry for webhook: %s" % e.columns
        )

    return webhook


@db_base.insecure_aware()
@db_base.session_aware()
def get_webhook(id, insecure=None, session=None):
    webhook = _get_db_object_by_id(models.Webhook, id, insecure=insecure)

    if not webhook:
        raise exc.DBEntityNotFoundError("Webhook not found [id=%s]" % id)

    return webhook


@db_base.session_aware()
def get_webhooks(session=None, **kwargs):
    return _get_collection_sorted_by_time(models.Webhook, **kwargs)


@db_base.session_aware()
def delete_webhook(id, session=None):
    webhook = get_webhook(id)
    session.delete(webhook)


@db_base.session_aware()
def update_webhook(id, values, session=None):
    webhook = get_webhook(id)
    webhook.update(values.copy())

    return webhook


@db_base.insecure_aware()
@db_base.session_aware()
def delete_webhooks(session=None, insecure=None, **kwargs):
    return _delete_all(models.Webhook, insecure=insecure, **kwargs)


@db_base.session_aware()
def increase_function_version(function_id, old_version, session=None,
                              **kwargs):
    """This function is supposed to be invoked within locking section."""
    version = models.FunctionVersion()
    kwargs.update(
        {
            "function_id": function_id,
            "version_number": old_version + 1
        }
    )
    version.update(kwargs.copy())

    try:
        version.save(session=session)
    except oslo_db_exc.DBDuplicateEntry as e:
        raise exc.DBError(
            "Duplicate entry for function_versions: %s" % e.columns
        )

    return version


@db_base.insecure_aware()
@db_base.session_aware()
def get_function_version(function_id, version, session=None, insecure=None):
    if insecure:
        query = db_base.model_query(models.FunctionVersion)
    else:
        query = _secure_query(models.FunctionVersion)

    version_db = query.filter_by(
        function_id=function_id, version_number=version
    ).first()

    if not version_db:
        raise exc.DBEntityNotFoundError(
            "FunctionVersion not found [function_id=%s, version_number=%s]" %
            (function_id, version)
        )

    return version_db


# This function is only used in unit test.
@db_base.session_aware()
def update_function_version(function_id, version, session=None, **kwargs):
    version_db = get_function_version(function_id, version, session=session)
    version_db.update(kwargs.copy())

    return version_db


@db_base.session_aware()
def delete_function_version(function_id, version, session=None):
    version_db = get_function_version(function_id, version)
    session.delete(version_db)


@db_base.session_aware()
def get_function_versions(session=None, **kwargs):
    return _get_collection_sorted_by_time(models.FunctionVersion, **kwargs)


@db_base.session_aware()
def create_function_alias(session=None, **kwargs):
    alias = models.FunctionAlias()
    alias.update(kwargs.copy())

    try:
        alias.save(session=session)
    except oslo_db_exc.DBDuplicateEntry as e:
        raise exc.DBError(
            "Duplicate entry for function_aliases: %s" % e.columns
        )

    return alias


@db_base.insecure_aware()
@db_base.session_aware()
def get_function_alias(name, session=None, insecure=None):
    alias = _get_db_object_by_name(models.FunctionAlias,
                                   name,
                                   insecure=insecure)

    if not alias:
        raise exc.DBEntityNotFoundError(
            "FunctionAlias not found [name=%s]" %
            (name)
        )

    return alias


@db_base.session_aware()
def get_function_aliases(session=None, **kwargs):
    return _get_collection_sorted_by_time(models.FunctionAlias, **kwargs)


@db_base.session_aware()
def update_function_alias(name, session=None, **kwargs):
    alias_db = get_function_alias(name, session=session)
    alias_db.update(kwargs.copy())

    return alias_db


@db_base.session_aware()
def delete_function_alias(name, session=None):
    alias_db = get_function_alias(name)
    session.delete(alias_db)


@db_base.session_aware()
def delete_function_aliases(session=None, **kwargs):
    return _delete_all(models.FunctionAlias, **kwargs)



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import abc

import six
from stevedore import driver

from qinling import exceptions as exc

STORAGE_PROVIDER = None


@six.add_metaclass(abc.ABCMeta)
class PackageStorage(object):
    """PackageStorage interface."""

    @abc.abstractmethod
    def store(self, project_id, function, data, **kwargs):
        """Store the function package data.

        :param project_id: Project ID.
        :param function: Function ID.
        :param data: Package file content.
        :param kwargs: A dict may including
            - md5sum: The MD5 provided by the user.
        :return: A tuple (if the package is updated, MD5 value of the package)
        """
        raise NotImplementedError

    @abc.abstractmethod
    def retrieve(self, project_id, function, md5sum, version=0):
        """Get function package data.

        :param project_id: Project ID.
        :param function: Function ID.
        :param md5sum: The function MD5.
        :param version: Optional. The function version number.
        :return: File descriptor that needs to close outside.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self, project_id, function, md5sum, version=0):
        raise NotImplementedError

    @abc.abstractmethod
    def changed_since(self, project_id, function, l_md5, version):
        """Check if the function package has changed.

        Check if the function package has changed between lastest and the
        specified version.

        :param project_id: Project ID.
        :param function: Function ID.
        :param l_md5: Latest function package md5sum.
        :param version: The version number compared with.
        :return: True if changed otherwise False.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def copy(self, project_id, function, l_md5, old_version):
        """Copy function package for a new version.

        :param project_id: Project ID.
        :param function: Function ID.
        :param l_md5: Latest function package md5sum.
        :param old_version: The version number that should copy from.
        :return: None
        """
        raise NotImplementedError


def load_storage_provider(conf):
    global STORAGE_PROVIDER

    if not STORAGE_PROVIDER:
        try:
            mgr = driver.DriverManager(
                'qinling.storage.provider',
                conf.storage.provider,
                invoke_on_load=True,
                invoke_args=[conf]
            )

            STORAGE_PROVIDER = mgr.driver
        except Exception as e:
            raise exc.StorageProviderException(
                'Failed to load storage provider: %s. Error: %s' %
                (conf.storage.provider, str(e))
            )

    return STORAGE_PROVIDER



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from datetime import datetime
import json
import tempfile
import uuid

import mock
from oslo_config import cfg

from qinling.db import api as db_api
from qinling import status
from qinling.tests.unit.api import base
from qinling.tests.unit import base as unit_base
from qinling.utils import constants


class TestFunctionController(base.APITest):
    def setUp(self):
        super(TestFunctionController, self).setUp()

        # Insert a runtime record in db for each test case. The data will be
        # removed automatically in tear down.
        db_runtime = self.create_runtime()
        self.runtime_id = db_runtime.id

    @mock.patch('qinling.storage.file_system.FileSystemStorage.store')
    def test_post(self, mock_store):
        mock_store.return_value = (True, 'fake_md5')

        with tempfile.NamedTemporaryFile() as f:
            body = {
                'name': self.rand_name('function', prefix=self.prefix),
                'code': json.dumps({"source": "package"}),
                'runtime_id': self.runtime_id,
            }
            resp = self.app.post(
                '/v1/functions',
                params=body,
                upload_files=[('package', f.name, f.read())]
            )

        self.assertEqual(201, resp.status_int)
        self.assertEqual(1, mock_store.call_count)

        body.update(
            {
                'entry': 'main.main',
                'code': {"source": "package", "md5sum": "fake_md5"},
                'timeout': cfg.CONF.resource_limits.default_timeout
            }
        )
        self._assertDictContainsSubset(resp.json, body)

    @mock.patch('qinling.storage.file_system.FileSystemStorage.store')
    def test_post_timeout(self, mock_store):
        mock_store.return_value = (True, 'fake_md5')

        with tempfile.NamedTemporaryFile() as f:
            body = {
                'runtime_id': self.runtime_id,
                'code': json.dumps({"source": "package"}),
                'timeout': 3
            }
            resp = self.app.post(
                '/v1/functions',
                params=body,
                upload_files=[('package', f.name, f.read())]
            )

        self.assertEqual(201, resp.status_int)
        self.assertEqual(3, resp.json['timeout'])

    def test_post_timeout_invalid(self):
        with tempfile.NamedTemporaryFile() as f:
            body = {
                'runtime_id': self.runtime_id,
                'code': json.dumps({"source": "package"}),
                'timeout': cfg.CONF.resource_limits.max_timeout + 1
            }
            resp = self.app.post(
                '/v1/functions',
                params=body,
                upload_files=[('package', f.name, f.read())],
                expect_errors=True
            )

        self.assertEqual(400, resp.status_int)
        self.assertIn(
            'timeout resource limitation not within the allowable range',
            resp.json['faultstring']
        )

    @mock.patch("qinling.utils.openstack.keystone.create_trust")
    @mock.patch('qinling.utils.openstack.keystone.get_swiftclient')
    @mock.patch('qinling.context.AuthHook.before')
    def test_post_from_swift(self, mock_auth, mock_client, mock_trust):
        self.override_config('auth_enable', True, group='pecan')

        swift_conn = mock.Mock()
        mock_client.return_value = swift_conn
        swift_conn.head_object.return_value = {
            'accept-ranges': 'bytes',
            'content-length': str(constants.MAX_PACKAGE_SIZE - 1)
        }
        mock_trust.return_value.id = str(uuid.uuid4())

        body = {
            'name': 'swift_function',
            'code': json.dumps(
                {
                    "source": "swift",
                    "swift": {"container": "container", "object": "object"}
                }
            ),
            'runtime_id': self.runtime_id,
        }
        resp = self.app.post('/v1/functions', params=body)

        self.assertEqual(201, resp.status_int)

        body.update(
            {
                'entry': 'main.main',
                'code': {
                    "source": "swift",
                    "swift": {"container": "container", "object": "object"}
                }
            }
        )
        self._assertDictContainsSubset(resp.json, body)

    def test_post_swift_not_enough_params(self):
        body = {
            'name': 'swift_function',
            'code': json.dumps(
                {
                    "source": "swift",
                    "swift": {"container": "fake-container"}
                }
            ),
            'runtime_id': self.runtime_id,
        }
        resp = self.app.post(
            '/v1/functions',
            params=body,
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)

    @mock.patch('qinling.utils.openstack.keystone.get_swiftclient')
    @mock.patch('qinling.context.AuthHook.before')
    def test_post_swift_size_exceed(self, mock_auth, mock_client):
        self.override_config('auth_enable', True, group='pecan')
        swift_conn = mock.Mock()
        mock_client.return_value = swift_conn
        swift_conn.head_object.return_value = {
            'accept-ranges': 'bytes',
            'content-length': str(constants.MAX_PACKAGE_SIZE + 1)
        }

        body = {
            'name': 'swift_function',
            'code': json.dumps(
                {
                    "source": "swift",
                    "swift": {"container": "container", "object": "object"}
                }
            ),
            'runtime_id': self.runtime_id,
        }
        resp = self.app.post(
            '/v1/functions',
            params=body,
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)

    def test_get(self):
        db_func = self.create_function(runtime_id=self.runtime_id)
        expected = {
            'id': db_func.id,
            "code": {"source": "package", "md5sum": "fake_md5"},
            "name": db_func.name,
            'entry': 'main.main',
            "project_id": unit_base.DEFAULT_PROJECT_ID,
            "cpu": cfg.CONF.resource_limits.default_cpu,
            "memory_size": cfg.CONF.resource_limits.default_memory,
            "timeout": cfg.CONF.resource_limits.default_timeout,
        }

        resp = self.app.get('/v1/functions/%s' % db_func.id)

        self.assertEqual(200, resp.status_int)
        self._assertDictContainsSubset(resp.json, expected)

    def test_get_all(self):
        db_func = self.create_function(runtime_id=self.runtime_id)
        expected = {
            'id': db_func.id,
            "name": db_func.name,
            'entry': 'main.main',
            "project_id": unit_base.DEFAULT_PROJECT_ID,
            "cpu": cfg.CONF.resource_limits.default_cpu,
            "memory_size": cfg.CONF.resource_limits.default_memory,
            "timeout": cfg.CONF.resource_limits.default_timeout,
        }

        resp = self.app.get('/v1/functions')

        self.assertEqual(200, resp.status_int)
        actual = self._assert_single_item(
            resp.json['functions'], id=db_func.id
        )
        self._assertDictContainsSubset(actual, expected)

    def test_put_name(self):
        db_func = self.create_function(runtime_id=self.runtime_id)

        resp = self.app.put_json(
            '/v1/functions/%s' % db_func.id, {'name': 'new_name'}
        )

        self.assertEqual(200, resp.status_int)
        self.assertEqual('new_name', resp.json['name'])

    def test_put_timeout(self):
        db_func = self.create_function(runtime_id=self.runtime_id)

        resp = self.app.put_json(
            '/v1/functions/%s' % db_func.id, {'timeout': 10}
        )

        self.assertEqual(200, resp.status_int)
        self.assertEqual(10, resp.json['timeout'])

    def test_put_timeout_invalid(self):
        db_func = self.create_function(runtime_id=self.runtime_id)

        # Check for type of cpu values.
        resp = self.app.put_json(
            '/v1/functions/%s' % db_func.id,
            {'timeout': cfg.CONF.resource_limits.max_timeout + 1},
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)
        self.assertIn(
            'timeout resource limitation not within the allowable range',
            resp.json['faultstring']
        )

    @mock.patch('qinling.utils.etcd_util.delete_function')
    @mock.patch('qinling.storage.file_system.FileSystemStorage.store')
    @mock.patch('qinling.storage.file_system.FileSystemStorage.delete')
    @mock.patch('qinling.rpc.EngineClient.delete_function')
    def test_put_package(self, mock_delete_func, mock_delete, mock_store,
                         mock_etcd_del):
        db_func = self.create_function(runtime_id=self.runtime_id)
        mock_store.return_value = (True, "fake_md5_changed")

        with tempfile.NamedTemporaryFile() as f:
            resp = self.app.put(
                '/v1/functions/%s' % db_func.id,
                params={},
                upload_files=[('package', f.name, f.read())]
            )

        self.assertEqual(200, resp.status_int)
        self.assertEqual(1, mock_store.call_count)
        self.assertEqual('fake_md5_changed', resp.json['code'].get('md5sum'))

        mock_delete_func.assert_called_once_with(db_func.id)
        mock_etcd_del.assert_called_once_with(db_func.id)
        mock_delete.assert_called_once_with(unit_base.DEFAULT_PROJECT_ID,
                                            db_func.id, "fake_md5")

    @mock.patch('qinling.storage.file_system.FileSystemStorage.store')
    @mock.patch('qinling.rpc.EngineClient.delete_function')
    @mock.patch('qinling.utils.etcd_util.delete_function')
    @mock.patch('qinling.storage.file_system.FileSystemStorage.delete')
    def test_put_package_md5_not_change(self, file_delete_mock,
                                        etcd_delete_mock, function_delete_mock,
                                        store_mock):
        db_func = self.create_function(runtime_id=self.runtime_id)
        store_mock.return_value = (False, "fake_md5")

        with tempfile.NamedTemporaryFile() as f:
            resp = self.app.put(
                '/v1/functions/%s' % db_func.id,
                params={},
                upload_files=[('package', f.name, f.read())]
            )

        self.assertEqual(200, resp.status_int)
        self.assertEqual('fake_md5', resp.json['code'].get('md5sum'))
        function_delete_mock.assert_called_once_with(db_func.id)
        etcd_delete_mock.assert_called_once_with(db_func.id)
        self.assertFalse(file_delete_mock.called)

    def test_put_package_same_md5_provided(self):
        db_func = self.create_function(runtime_id=self.runtime_id)

        with tempfile.NamedTemporaryFile() as f:
            resp = self.app.put(
                '/v1/functions/%s' % db_func.id,
                params={
                    "code": json.dumps(
                        {"md5sum": "fake_md5", "source": "package"}
                    )
                },
                upload_files=[('package', f.name, f.read())],
                expect_errors=True
            )

        self.assertEqual(400, resp.status_int)

    @mock.patch('qinling.rpc.EngineClient.delete_function')
    @mock.patch('qinling.utils.etcd_util.delete_function')
    @mock.patch('qinling.utils.openstack.swift.check_object')
    @mock.patch('qinling.context.AuthHook.before')
    def test_put_swift_function(self, mock_auth, mock_check, mock_etcd_delete,
                                mock_func_delete):
        self.override_config('auth_enable', True, group='pecan')
        mock_check.return_value = True

        db_func = self.create_function(
            runtime_id=self.runtime_id,
            code={
                "source": "swift",
                "swift": {"container": "fake-container", "object": "fake-obj"}
            }
        )

        body = {
            'code': json.dumps(
                {
                    "source": "swift",
                    "swift": {"object": "new-obj"}
                }
            ),
        }
        resp = self.app.put_json('/v1/functions/%s' % db_func.id, body)

        self.assertEqual(200, resp.status_int)
        swift_info = {
            'code': {
                "source": "swift",
                "swift": {"container": "fake-container", "object": "new-obj"}
            }
        }
        self._assertDictContainsSubset(resp.json, swift_info)

    @mock.patch('qinling.rpc.EngineClient.delete_function')
    @mock.patch('qinling.utils.etcd_util.delete_function')
    @mock.patch('qinling.utils.openstack.swift.check_object')
    @mock.patch('qinling.context.AuthHook.before')
    def test_put_swift_function_without_source(self, mock_auth, mock_check,
                                               mock_etcd_delete,
                                               mock_func_delete):
        self.override_config('auth_enable', True, group='pecan')
        mock_check.return_value = True

        db_func = self.create_function(
            runtime_id=self.runtime_id,
            code={
                "source": "swift",
                "swift": {"container": "fake-container", "object": "fake-obj"}
            }
        )

        body = {
            'code': json.dumps(
                {
                    "swift": {"object": "new-obj"}
                }
            ),
        }
        resp = self.app.put_json('/v1/functions/%s' % db_func.id, body)

        self.assertEqual(200, resp.status_int)
        swift_info = {
            'code': {
                "source": "swift",
                "swift": {"container": "fake-container", "object": "new-obj"}
            }
        }
        self._assertDictContainsSubset(resp.json, swift_info)

    def test_put_cpu_with_type_error(self):
        db_func = self.create_function(runtime_id=self.runtime_id)

        # Check for type of cpu values.
        resp = self.app.put_json(
            '/v1/functions/%s' % db_func.id, {'cpu': 'non-int'},
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)
        self.assertIn(
            'Invalid cpu resource specified. An integer is required.',
            resp.json['faultstring']
        )

    def test_put_cpu_with_overrun_error(self):
        db_func = self.create_function(runtime_id=self.runtime_id)

        # Check for cpu error with input out of range.
        resp = self.app.put_json(
            '/v1/functions/%s' % db_func.id, {'cpu': 0},
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)
        self.assertIn(
            'cpu resource limitation not within the allowable range',
            resp.json['faultstring']
        )

    @mock.patch('qinling.storage.file_system.FileSystemStorage.delete')
    @mock.patch('qinling.utils.etcd_util.delete_function')
    @mock.patch('qinling.rpc.EngineClient.delete_function')
    def test_put_cpu_and_memorysize(self, mock_delete_func, mock_etcd_del,
                                    mock_storage_delete):
        # Test for updating cpu/mem with good input values.
        db_func = self.create_function(runtime_id=self.runtime_id)

        req_body = {
            'cpu': str(cfg.CONF.resource_limits.default_cpu + 1),
            'memory_size': str(cfg.CONF.resource_limits.default_memory + 1)
        }

        resp = self.app.put_json('/v1/functions/%s' % db_func.id, req_body)

        self.assertEqual(200, resp.status_int)
        self.assertEqual(
            cfg.CONF.resource_limits.default_cpu + 1,
            resp.json['cpu']
        )
        self.assertEqual(
            cfg.CONF.resource_limits.default_memory + 1,
            resp.json['memory_size']
        )
        mock_delete_func.assert_called_once_with(db_func.id)
        mock_etcd_del.assert_called_once_with(db_func.id)
        self.assertFalse(mock_storage_delete.called)

    @mock.patch('qinling.utils.etcd_util.delete_function')
    @mock.patch('qinling.rpc.EngineClient.delete_function')
    @mock.patch('qinling.storage.file_system.FileSystemStorage.delete')
    def test_delete(self, mock_delete, mock_delete_func, mock_etcd_delete):
        db_func = self.create_function(runtime_id=self.runtime_id)
        resp = self.app.delete('/v1/functions/%s' % db_func.id)

        self.assertEqual(204, resp.status_int)
        mock_delete.assert_called_once_with(
            unit_base.DEFAULT_PROJECT_ID, db_func.id, "fake_md5"
        )
        mock_delete_func.assert_called_once_with(db_func.id)
        mock_etcd_delete.assert_called_once_with(db_func.id)

    def test_delete_with_running_job(self):
        db_func = self.create_function(runtime_id=self.runtime_id)
        self.create_job(
            function_id=db_func.id,
            status=status.AVAILABLE,
            first_execution_time=datetime.utcnow(),
            next_execution_time=datetime.utcnow(),
            count=1
        )

        resp = self.app.delete(
            '/v1/functions/%s' % db_func.id,
            expect_errors=True
        )

        self.assertEqual(403, resp.status_int)

    def test_delete_with_webhook(self):
        db_func = self.create_function(runtime_id=self.runtime_id)
        self.create_webhook(function_id=db_func.id)

        resp = self.app.delete(
            '/v1/functions/%s' % db_func.id,
            expect_errors=True
        )

        self.assertEqual(403, resp.status_int)

    @mock.patch('qinling.utils.etcd_util.delete_function')
    @mock.patch('qinling.rpc.EngineClient.delete_function')
    @mock.patch('qinling.storage.file_system.FileSystemStorage.delete')
    def test_delete_with_versions(self, mock_package_delete,
                                  mock_engine_delete, mock_etcd_delete):
        db_func = self.create_function(runtime_id=self.runtime_id)
        func_id = db_func.id
        # Create two versions for the function
        db_api.increase_function_version(func_id, 0)
        db_api.increase_function_version(func_id, 1)

        resp = self.app.delete('/v1/functions/%s' % func_id)

        self.assertEqual(204, resp.status_int)

        self.assertEqual(3, mock_package_delete.call_count)
        self.assertEqual(3, mock_engine_delete.call_count)
        self.assertEqual(3, mock_etcd_delete.call_count)

        mock_package_delete.assert_has_calls(
            [
                mock.call(unit_base.DEFAULT_PROJECT_ID, func_id, None,
                          version=1),
                mock.call(unit_base.DEFAULT_PROJECT_ID, func_id, None,
                          version=2),
                mock.call(unit_base.DEFAULT_PROJECT_ID, func_id, "fake_md5")
            ]
        )

        mock_engine_delete.assert_has_calls(
            [
                mock.call(func_id, version=1),
                mock.call(func_id, version=2),
                mock.call(func_id)
            ]
        )

        mock_etcd_delete.assert_has_calls(
            [
                mock.call(func_id, version=1),
                mock.call(func_id, version=2),
                mock.call(func_id)
            ]
        )

    def test_delete_with_version_associate_webhook(self):
        db_func = self.create_function(runtime_id=self.runtime_id)
        func_id = db_func.id
        db_api.increase_function_version(func_id, 0)
        self.create_webhook(func_id, function_version=1)

        resp = self.app.delete(
            '/v1/functions/%s' % func_id,
            expect_errors=True
        )

        self.assertEqual(403, resp.status_int)

    def test_delete_with_alias(self):
        db_func = self.create_function(runtime_id=self.runtime_id)
        func_id = db_func.id
        name = self.rand_name(name="alias", prefix=self.prefix)
        body = {
            'function_id': func_id,
            'name': name
        }
        db_api.create_function_alias(**body)

        resp = self.app.delete(
            '/v1/functions/%s' % func_id,
            expect_errors=True
        )

        self.assertEqual(403, resp.status_int)

    @mock.patch('qinling.rpc.EngineClient.scaleup_function')
    def test_scale_up(self, scaleup_function_mock):
        db_func = self.create_function(runtime_id=self.runtime_id)

        body = {'count': 1}
        resp = self.app.post(
            '/v1/functions/%s/scale_up' % db_func.id,
            params=json.dumps(body),
            content_type='application/json'
        )

        self.assertEqual(202, resp.status_int)
        scaleup_function_mock.assert_called_once_with(
            db_func.id, runtime_id=self.runtime_id, count=1)

    @mock.patch('qinling.utils.etcd_util.get_workers')
    @mock.patch('qinling.rpc.EngineClient.scaledown_function')
    def test_scale_down(self, scaledown_function_mock, get_workers_mock):
        db_func = self.create_function(runtime_id=self.runtime_id)
        get_workers_mock.return_value = [mock.Mock(), mock.Mock()]

        body = {'count': 1}
        resp = self.app.post(
            '/v1/functions/%s/scale_down' % db_func.id,
            params=json.dumps(body),
            content_type='application/json'
        )

        self.assertEqual(202, resp.status_int)
        scaledown_function_mock.assert_called_once_with(db_func.id, count=1)

    @mock.patch('qinling.utils.etcd_util.get_workers')
    @mock.patch('qinling.rpc.EngineClient.scaledown_function')
    def test_scale_down_no_need(
            self, scaledown_function_mock, get_workers_mock
    ):
        db_func = self.create_function(runtime_id=self.runtime_id)
        get_workers_mock.return_value = [mock.Mock()]

        body = {'count': 1}
        resp = self.app.post(
            '/v1/functions/%s/scale_down' % db_func.id,
            params=json.dumps(body),
            content_type='application/json'
        )

        self.assertEqual(202, resp.status_int)
        scaledown_function_mock.assert_not_called()

    @mock.patch('qinling.utils.etcd_util.delete_function')
    @mock.patch('qinling.rpc.EngineClient.delete_function')
    def test_detach(
            self, engine_delete_function_mock, etcd_delete_function_mock
    ):
        db_func = self.create_function(runtime_id=self.runtime_id)

        resp = self.app.post(
            '/v1/functions/%s/detach' % db_func.id
        )

        self.assertEqual(202, resp.status_int)
        engine_delete_function_mock.assert_called_once_with(db_func.id)
        etcd_delete_function_mock.assert_called_once_with(db_func.id)



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_config import cfg
from oslo_log import log as logging
import requests
import tenacity

from qinling.db import api as db_api
from qinling.engine import utils
from qinling import exceptions as exc
from qinling import status
from qinling.utils import constants
from qinling.utils import etcd_util

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class DefaultEngine(object):
    def __init__(self, orchestrator, qinling_endpoint):
        self.orchestrator = orchestrator
        self.qinling_endpoint = qinling_endpoint
        self.session = requests.Session()

    def create_runtime(self, ctx, runtime_id):
        LOG.info('Start to create runtime %s.', runtime_id)

        with db_api.transaction():
            runtime = db_api.get_runtime(runtime_id)

            try:
                self.orchestrator.create_pool(
                    runtime_id,
                    runtime.image,
                    trusted=runtime.trusted
                )
                runtime.status = status.AVAILABLE
                LOG.info('Runtime %s created.', runtime_id)
            except Exception as e:
                LOG.exception(
                    'Failed to create pool for runtime %s. Error: %s',
                    runtime_id,
                    str(e)
                )
                runtime.status = status.ERROR

    def delete_runtime(self, ctx, runtime_id):
        LOG.info('Start to delete runtime %s.', runtime_id)

        self.orchestrator.delete_pool(runtime_id)
        db_api.delete_runtime(runtime_id)

        LOG.info('Deleted runtime %s.', runtime_id)

    def update_runtime(self, ctx, runtime_id, image=None, pre_image=None):
        LOG.info(
            'Start to update runtime %s, image: %s, pre_image: %s',
            runtime_id, image, pre_image
        )

        ret = self.orchestrator.update_pool(runtime_id, image=image)

        if ret:
            values = {'status': status.AVAILABLE}
            db_api.update_runtime(runtime_id, values)

            LOG.info('Updated runtime %s.', runtime_id)
        else:
            values = {'status': status.AVAILABLE, 'image': pre_image}
            db_api.update_runtime(runtime_id, values)

            LOG.info('Rollbacked runtime %s.', runtime_id)

    def get_runtime_pool(self, ctx, runtime_id):
        LOG.info("Getting pool information for runtime %s", runtime_id)

        return self.orchestrator.get_pool(runtime_id)

    @tenacity.retry(
        wait=tenacity.wait_fixed(1),
        stop=tenacity.stop_after_attempt(30),
        reraise=True,
        retry=tenacity.retry_if_exception_type(exc.EtcdLockException)
    )
    def function_load_check(self, function_id, version, runtime_id):
        """Check function load and scale the workers if needed.

        :return: None if no need to scale up otherwise return the service url
        """
        with etcd_util.get_worker_lock(function_id, version) as lock:
            if not lock.is_acquired():
                raise exc.EtcdLockException(
                    'Etcd: failed to get worker lock for function %s'
                    '(version %s).' % (function_id, version)
                )

            workers = etcd_util.get_workers(function_id, version)
            running_execs = db_api.get_executions(
                function_id=function_id,
                function_version=version,
                status=status.RUNNING
            )
            concurrency = (len(running_execs) or 1) / (len(workers) or 1)
            if (len(workers) == 0 or
                    concurrency > CONF.engine.function_concurrency):
                LOG.info(
                    'Scale up function %s(version %s). Current concurrency: '
                    '%s, execution number %s, worker number %s',
                    function_id, version, concurrency, len(running_execs),
                    len(workers)
                )

                # NOTE(kong): The increase step could be configurable
                return self.scaleup_function(None, function_id, version,
                                             runtime_id, 1)

    def create_execution(self, ctx, execution_id, function_id,
                         function_version, runtime_id, input=None):
        LOG.info(
            'Creating execution. execution_id=%s, function_id=%s, '
            'function_version=%s, runtime_id=%s, input=%s',
            execution_id, function_id, function_version, runtime_id, input
        )

        function = db_api.get_function(function_id)
        source = function.code['source']
        rlimit = {
            'cpu': function.cpu,
            'memory_size': function.memory_size
        }
        image = None
        identifier = None
        labels = None
        svc_url = None
        is_image_source = source == constants.IMAGE_FUNCTION

        # Auto scale workers if needed
        if not is_image_source:
            try:
                svc_url = self.function_load_check(function_id,
                                                   function_version,
                                                   runtime_id)
            except (
                exc.OrchestratorException,
                exc.EtcdLockException
            ) as e:
                utils.handle_execution_exception(execution_id, str(e))
                return

        temp_url = etcd_util.get_service_url(function_id, function_version)
        svc_url = svc_url or temp_url
        if svc_url:
            func_url = '%s/execute' % svc_url
            LOG.debug(
                'Found service url for function: %s(version %s), execution: '
                '%s, url: %s',
                function_id, function_version, execution_id, func_url
            )

            data = utils.get_request_data(
                CONF, function_id, function_version, execution_id,
                rlimit, input, function.entry, function.trust_id,
                self.qinling_endpoint, function.timeout
            )
            success, res = utils.url_request(
                self.session, func_url, body=data
            )

            utils.finish_execution(execution_id, success, res,
                                   is_image_source=is_image_source)
            return

        if is_image_source:
            image = function.code['image']
            identifier = ('%s-%s' % (execution_id, function_id))[:63]
        else:
            identifier = runtime_id
            labels = {'runtime_id': runtime_id}

        try:
            # For image function, it will be executed inside this method;
            # For package type function it only sets up underlying resources
            # and get a service url. If the service url is already created
            # beforehand, nothing happens.
            _, svc_url = self.orchestrator.prepare_execution(
                function_id,
                function_version,
                rlimit=rlimit,
                image=image,
                identifier=identifier,
                labels=labels,
                input=input,
            )
        except exc.OrchestratorException as e:
            utils.handle_execution_exception(execution_id, str(e))
            return

        # For image type function, wait for its completion and retrieve the
        # worker log;
        # For package type function, invoke and get log
        success, res = self.orchestrator.run_execution(
            execution_id,
            function_id,
            function_version,
            rlimit=rlimit if svc_url else None,
            input=input,
            identifier=identifier,
            service_url=svc_url,
            entry=function.entry,
            trust_id=function.trust_id,
            timeout=function.timeout
        )

        utils.finish_execution(execution_id, success, res,
                               is_image_source=is_image_source)

    def delete_function(self, ctx, function_id, function_version=0):
        """Deletes underlying resources allocated for function."""
        LOG.info('Start to delete function %s(version %s).', function_id,
                 function_version)

        self.orchestrator.delete_function(function_id, function_version)

        LOG.info('Deleted function %s(version %s).', function_id,
                 function_version)

    def scaleup_function(self, ctx, function_id, function_version, runtime_id,
                         count=1):
        worker_names, service_url = self.orchestrator.scaleup_function(
            function_id,
            function_version,
            identifier=runtime_id,
            count=count
        )

        for name in worker_names:
            etcd_util.create_worker(function_id, name,
                                    version=function_version)

        etcd_util.create_service_url(function_id, service_url,
                                     version=function_version)

        LOG.info('Finished scaling up function %s(version %s).', function_id,
                 function_version)

        return service_url

    def scaledown_function(self, ctx, function_id, function_version=0,
                           count=1):
        workers = etcd_util.get_workers(function_id, function_version)
        worker_deleted_num = (
            count if len(workers) > count else len(workers) - 1
        )
        workers = workers[:worker_deleted_num]

        for worker in workers:
            LOG.debug('Removing worker %s', worker)
            self.orchestrator.delete_worker(worker)
            etcd_util.delete_worker(function_id, worker,
                                    version=function_version)

        LOG.info('Finished scaling down function %s(version %s).', function_id,
                 function_version)



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


class QinlingException(Exception):
    """Qinling specific exception.

    Reserved for situations that are not critical for program continuation.
    It is possible to recover from this type of problems automatically and
    continue program execution. Such problems may be related with invalid user
    input (such as invalid syntax) or temporary environmental problems.

    In case if an instance of a certain exception type bubbles up to API layer
    then this type of exception it must be associated with an http code so it's
    clear how to represent it for a client.

    To correctly use this class, inherit from it and define a 'message' and
    'http_code' properties.
    """
    message = "An unknown exception occurred"
    http_code = 500

    def __init__(self, message=None):
        if message is not None:
            self.message = message

        super(QinlingException, self).__init__(
            '%d: %s' % (self.http_code, self.message))

    @property
    def code(self):
        """This is here for webob to read.

        https://github.com/Pylons/webob/blob/master/webob/exc.py
        """
        return self.http_code

    def __str__(self):
        return self.message


class InputException(QinlingException):
    http_code = 400


class UnauthorizedException(QinlingException):
    http_code = 401
    message = "Unauthorized"


class NotAllowedException(QinlingException):
    http_code = 403
    message = "Operation not allowed"


class ConflictException(QinlingException):
    http_code = 409
    message = ("The request could not be completed due to a conflict with the "
               "current state of the target resource")


class RuntimeNotAvailableException(QinlingException):
    http_code = 409
    message = "Runtime not available"


class DBError(QinlingException):
    http_code = 400


class DBEntityNotFoundError(DBError):
    http_code = 404
    message = "Object not found"


class RuntimeNotFoundException(QinlingException):
    http_code = 404
    message = "Runtime not found"


class ApplicationContextNotFoundException(QinlingException):
    http_code = 400
    message = "Application context not found"


class StorageNotFoundException(QinlingException):
    http_code = 404
    message = "Storage file not found"


class StorageProviderException(QinlingException):
    http_code = 500


class OrchestratorException(QinlingException):
    http_code = 500
    message = "Orchestrator error."


class TrustFailedException(QinlingException):
    http_code = 500
    message = "Trust operation failed."


class SwiftException(QinlingException):
    http_code = 500
    message = "Failed to communicate with Swift."


class EtcdLockException(QinlingException):
    http_code = 409
    message = 'Etcd lock failed'


class TimeoutException(QinlingException):
    http_code = 500
    message = 'Function execution timeout'



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from oslo_log import log as logging
from oslo_serialization import jsonutils

from qinling.db import api as db_api
from qinling.db.sqlalchemy import models
from qinling import exceptions as exc
from qinling import status
from qinling.utils import constants

LOG = logging.getLogger(__name__)


def _update_function_db(function_id, pre_count):
    # Function update is done using UPDATE ... FROM ... WHERE
    # non-locking clause.
    while True:
        modified = db_api.conditional_update(
            models.Function,
            {
                'count': pre_count + 1,
            },
            {
                'id': function_id,
                'count': pre_count
            },
            insecure=True,
        )
        if not modified:
            LOG.warning("Retrying to update function count.")
            pre_count += 1
            continue
        else:
            break


def _update_function_version_db(version_id, pre_count):
    # Update is done using UPDATE ... FROM ... WHERE non-locking clause.
    while True:
        modified = db_api.conditional_update(
            models.FunctionVersion,
            {
                'count': pre_count + 1,
            },
            {
                'id': version_id,
                'count': pre_count
            },
            insecure=True,
        )
        if not modified:
            LOG.warning("Retrying to update function version count.")
            pre_count += 1
            continue
        else:
            break


def create_execution(engine_client, params):
    function_alias = params.get('function_alias')
    function_id = params.get('function_id')
    version = params.get('function_version', 0)
    is_sync = params.get('sync', True)
    input = params.get('input')

    if function_alias:
        alias_db = db_api.get_function_alias(function_alias)
        function_id = alias_db.function_id
        version = alias_db.function_version
        params.update({'function_id': function_id,
                       'function_version': version})

    func_db = db_api.get_function(function_id)
    runtime_id = func_db.runtime_id

    # Image type function does not need runtime
    if runtime_id:
        runtime_db = db_api.get_runtime(runtime_id)
        if runtime_db and runtime_db.status != status.AVAILABLE:
            raise exc.RuntimeNotAvailableException(
                'Runtime %s is not available.' % func_db.runtime_id
            )

    if version > 0:
        if func_db.code['source'] != constants.PACKAGE_FUNCTION:
            raise exc.InputException(
                "Can not specify version for %s type function." %
                constants.PACKAGE_FUNCTION
            )

        # update version count
        version_db = db_api.get_function_version(function_id, version)
        pre_version_count = version_db.count
        _update_function_version_db(version_db.id, pre_version_count)
    else:
        pre_count = func_db.count
        _update_function_db(function_id, pre_count)

    # input in params should be a string.
    if input:
        try:
            function_input = jsonutils.loads(input)
            # If input is e.g. '6', result of jsonutils.loads is 6 which can
            # not be stored in db.
            if type(function_input) == int:
                raise ValueError
            params['input'] = function_input
        except ValueError:
            params['input'] = {'__function_input': input}

    params.update({'status': status.RUNNING})
    db_model = db_api.create_execution(params)

    try:
        engine_client.create_execution(
            db_model.id, function_id, version, runtime_id,
            input=params.get('input'), is_sync=is_sync
        )
    except exc.QinlingException:
        # Catch RPC errors for executions:
        #   - for RemoteError in an RPC call, the execution status would be
        #     handled in the engine side;
        #   - for other exceptions in an RPC call or cast, the execution status
        #     would remain RUNNING so we should update it.
        db_model = db_api.get_execution(db_model.id)
        if db_model.status == status.RUNNING:
            db_model = db_api.update_execution(db_model.id,
                                               {'status': status.ERROR})
        return db_model

    if is_sync:
        # The execution should already be updated by engine service for sync
        # execution.
        db_model = db_api.get_execution(db_model.id)

    return db_model



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_log import log as logging
from swiftclient.exceptions import ClientException

from qinling import exceptions as exc
from qinling.utils import common
from qinling.utils import constants
from qinling.utils.openstack import keystone

LOG = logging.getLogger(__name__)


@common.disable_ssl_warnings
def check_object(container, object):
    """Check object in Swift.

    1. If the object exists.
    2. Object size.

    :param container: Container name.
    :param object: Object name.
    :return: True if object exists, otherwise return False.
    """
    swift_conn = keystone.get_swiftclient()

    try:
        header = swift_conn.head_object(container, object)
    except ClientException:
        LOG.error(
            'The object %s in container %s was not found', object, container
        )
        return False
    except Exception:
        LOG.exception("Error when communicating with Swift.")
        return False

    if int(header['content-length']) > constants.MAX_PACKAGE_SIZE:
        LOG.error('Object size is greater than %s', constants.MAX_PACKAGE_SIZE)
        return False

    return True


@common.disable_ssl_warnings
def download_object(container, object):
    swift_conn = keystone.get_swiftclient()

    try:
        # Specify 'resp_chunk_size' here to return a file reader.
        _, obj_reader = swift_conn.get_object(
            container, object, resp_chunk_size=65536
        )
    except Exception:
        LOG.exception("Error when downloading object from Swift.")
        raise exc.SwiftException()

    return obj_reader



# Copyright 2015
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import os

from tempest import config
from tempest.test_discover import plugins

from qinling_tempest_plugin import config as qinling_config


class QinlingTempestPlugin(plugins.TempestPlugin):
    def load_tests(self):
        base_path = os.path.split(os.path.dirname(
            os.path.abspath(__file__))
        )[0]
        test_dir = "qinling_tempest_plugin/tests"
        full_test_dir = os.path.join(base_path, test_dir)
        return full_test_dir, base_path

    def register_opts(self, conf):
        conf.register_opt(
            qinling_config.service_option, group='service_available'
        )

        conf.register_group(qinling_config.qinling_group)
        conf.register_opts(qinling_config.QinlingGroup, group='qinling')

    def get_opt_lists(self):
        return [
            ('service_available', [qinling_config.service_option]),
            (qinling_config.qinling_group.name, qinling_config.QinlingGroup)
        ]

    def get_service_clients(self):
        qinling_config = config.service_client_config('qinling')
        params = {
            'name': 'qinling',
            'service_version': 'qinling',
            'module_path': 'qinling_tempest_plugin.services.qinling_client',
            'client_names': ['QinlingClient'],
        }
        params.update(qinling_config)
        return [params]



# Copyright 2018 OpenStack Foundation.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""add trusted field for runtimes table

Revision ID: 005
Revises: 004
Create Date: 2018-07-24 12:00:00.888969

"""

# revision identifiers, used by Alembic.
revision = '005'
down_revision = '004'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        'runtimes',
        sa.Column('trusted', sa.BOOLEAN, nullable=False, default=True,
                  server_default="1")
    )



# -*- coding: utf-8 -*-

# Copyright 2010-2011 OpenStack Foundation
# Copyright (c) 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import random

from oslo_config import cfg
from oslotest import base

from qinling import config
from qinling import context as auth_context
from qinling.db import api as db_api
from qinling import status

DEFAULT_PROJECT_ID = 'default'
OPT_PROJECT_ID = '55-66-77-88'


def get_context(default=True, admin=False):
    if default:
        return auth_context.Context.from_dict({
            'user_name': 'test-default-user',
            'user': '1-2-3-4',
            'tenant': DEFAULT_PROJECT_ID,
            'project_name': 'test-default-project',
            'is_admin': admin
        })
    else:
        return auth_context.Context.from_dict({
            'user_name': 'test-opt-user',
            'user': '5-6-7-8',
            'tenant': OPT_PROJECT_ID,
            'project_name': 'test-opt-project',
            'is_admin': admin
        })


class BaseTest(base.BaseTestCase):
    def override_config(self, name, override, group=None):
        """Cleanly override CONF variables."""
        cfg.CONF.set_override(name, override, group)
        self.addCleanup(cfg.CONF.clear_override, name, group)

    def _assertDictContainsSubset(self, parent, child):
        """Checks whether child dict is a superset of parent.

        assertDictContainsSubset() in standard Python 2.7 has been deprecated
        since Python 3.2

        Refer to https://goo.gl/iABb5c
        """
        self.assertEqual(parent, dict(parent, **child))

    def _assert_single_item(self, items, **props):
        return self._assert_multiple_items(items, 1, **props)[0]

    def _assert_multiple_items(self, items, count, **props):
        def _matches(item, **props):
            for prop_name, prop_val in props.items():
                v = (item[prop_name] if isinstance(item, dict)
                     else getattr(item, prop_name))
                if v != prop_val:
                    return False
            return True

        filtered_items = list(
            [item for item in items if _matches(item, **props)]
        )
        found = len(filtered_items)

        if found != count:
            self.fail("Wrong number of items found [props=%s, "
                      "expected=%s, found=%s]" % (props, count, found))

        return filtered_items

    def rand_name(self, name='', prefix=None):
        """Generate a random name that inclues a random number.

        :param str name: The name that you want to include
        :param str prefix: The prefix that you want to include
        :return: a random name. The format is
                 '<prefix>-<name>-<random number>'.
                 (e.g. 'prefixfoo-namebar-154876201')
        :rtype: string
        """
        randbits = str(random.randint(1, 0x7fffffff))
        rand_name = randbits
        if name:
            rand_name = name + '-' + rand_name
        if prefix:
            rand_name = prefix + '-' + rand_name
        return rand_name


class DbTestCase(BaseTest):
    is_heavy_init_called = False

    def setUp(self):
        super(DbTestCase, self).setUp()

        self.prefix = self.__class__.__name__

        self._heavy_init()

        self.ctx = get_context()
        auth_context.set_ctx(self.ctx)

        self.addCleanup(auth_context.set_ctx, None)
        self.addCleanup(self._clean_db)

    @classmethod
    def heavy_init(cls):
        """Runs a long initialization.

        This method runs long initialization once by class
        and can be extended by child classes.
        """
        cfg.CONF.set_default('connection', 'sqlite://', group='database')
        cfg.CONF.set_default('max_overflow', -1, group='database')
        cfg.CONF.set_default('max_pool_size', 1000, group='database')

        qinling_opts = [
            (config.API_GROUP, config.api_opts),
            (config.PECAN_GROUP, config.pecan_opts),
            (config.ENGINE_GROUP, config.engine_opts),
            (config.STORAGE_GROUP, config.storage_opts),
            (config.KUBERNETES_GROUP, config.kubernetes_opts),
            (config.ETCD_GROUP, config.etcd_opts),
            (config.RLIMITS_GROUP, config.rlimits_opts),
            (None, [config.launch_opt]),
            (None, config.default_opts)
        ]
        for group, options in qinling_opts:
            cfg.CONF.register_opts(list(options), group)
        cls.qinling_endpoint = 'http://127.0.0.1:7070/'
        cfg.CONF.set_default('qinling_endpoint', cls.qinling_endpoint)

        db_api.setup_db()

    @classmethod
    def _heavy_init(cls):
        """Method that runs heavy_init().

        Make this method private to prevent extending this one.
        It runs heavy_init() only once.

        Note: setUpClass() can be used, but it magically is not invoked
        from child class in another module.
        """
        if not cls.is_heavy_init_called:
            cls.heavy_init()
            cls.is_heavy_init_called = True

    def _clean_db(self):
        db_api.delete_all()

    def create_runtime(self):
        runtime = db_api.create_runtime(
            {
                'name': self.rand_name('runtime', prefix=self.prefix),
                'image': self.rand_name('image', prefix=self.prefix),
                # 'auth_enable' is disabled by default, we create runtime for
                # default tenant.
                'project_id': DEFAULT_PROJECT_ID,
                'status': status.AVAILABLE,
                'trusted': True
            }
        )

        return runtime

    def create_function(self, runtime_id=None, code=None, timeout=None):
        if not runtime_id:
            runtime_id = self.create_runtime().id

        function = db_api.create_function(
            {
                'name': self.rand_name('function', prefix=self.prefix),
                'runtime_id': runtime_id,
                'code': code or {"source": "package", "md5sum": "fake_md5"},
                'entry': 'main.main',
                # 'auth_enable' is disabled by default, we create runtime for
                # default tenant.
                'project_id': DEFAULT_PROJECT_ID,
                'cpu': cfg.CONF.resource_limits.default_cpu,
                'memory_size': cfg.CONF.resource_limits.default_memory,
                'timeout': timeout or cfg.CONF.resource_limits.default_timeout
            }
        )

        return function

    def create_job(self, function_id=None, function_alias=None, **kwargs):
        if not function_id and not function_alias:
            function_id = self.create_function().id

        job_params = {
            'name': self.rand_name('job', prefix=self.prefix),
            'function_alias': function_alias,
            'function_id': function_id,
            # 'auth_enable' is disabled by default
            'project_id': DEFAULT_PROJECT_ID,
        }
        job_params.update(kwargs)
        job = db_api.create_job(job_params)

        return job

    def create_webhook(self, function_id=None, function_alias=None, **kwargs):
        if not function_id and not function_alias:
            function_id = self.create_function().id

        webhook_params = {
            'function_alias': function_alias,
            'function_id': function_id,
            # 'auth_enable' is disabled by default
            'project_id': DEFAULT_PROJECT_ID,
        }
        webhook_params.update(kwargs)
        webhook = db_api.create_webhook(webhook_params)

        return webhook

    def create_execution(self, function_id=None, function_alias=None,
                         **kwargs):
        if not function_id and not function_alias:
            function_id = self.create_function().id

        execution_params = {
            'function_alias': function_alias,
            'function_id': function_id,
            'project_id': DEFAULT_PROJECT_ID,
            'status': status.RUNNING,
        }
        execution_params.update(kwargs)
        execution = db_api.create_execution(execution_params)

        return execution

    def create_function_version(self, old_version, function_id=None, **kwargs):
        if not function_id:
            function_id = self.create_function().id

        db_api.increase_function_version(function_id, old_version, **kwargs)
        db_api.update_function(function_id,
                               {"latest_version": old_version + 1})



# Copyright 2019 - Ormuco Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""add function_alias field for executions table
Revision ID: 008
Revises: 007
"""

revision = '008'
down_revision = '007'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        'executions',
        sa.Column('function_alias', sa.String(length=255), nullable=True)
    )



# Copyright 2018 AWCloud Software Co., Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


# Codes are from https://www.craig-wood.com/nick/articles/pi-machin/

def arctan_euler(x, one):
    x_squared = x * x
    x_squared_plus_1 = x_squared + 1
    term = (x * one) // x_squared_plus_1
    total = term
    two_n = 2
    while True:
        divisor = (two_n + 1) * x_squared_plus_1
        term *= two_n
        term = term // divisor
        if term == 0:
            break
        total += term
        two_n += 2
    return total


def pi_machin(one):
    return 4 * (4 * arctan_euler(5, one) - arctan_euler(239, one))


def main(digit=50000, *args, **kwargs):
    return str(pi_machin(10**digit))[:15]



# Copyright 2018 OpenStack Foundation.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from qinling import context
from qinling.db import api as db_api
from qinling import exceptions as exc
from qinling.tests.unit.api import base
from qinling.tests.unit import base as unit_base


class TestFunctionAliasController(base.APITest):
    def setUp(self):
        super(TestFunctionAliasController, self).setUp()

        self.db_func = self.create_function()
        self.func_id = self.db_func.id

    def test_post(self):
        name = 'TestAlias'
        body = {'function_id': self.func_id,
                'name': name,
                'description': 'new alias'}

        resp = self.app.post_json('/v1/aliases', body)

        self.assertEqual(201, resp.status_int)
        self._assertDictContainsSubset(resp.json, body)

        context.set_ctx(self.ctx)

        func_alias_db = db_api.get_function_alias(name)
        self.assertEqual(name, func_alias_db.name)
        self.assertEqual(0, func_alias_db.function_version)

    def test_post_without_required_params(self):
        body = {}

        resp = self.app.post_json('/v1/aliases',
                                  body,
                                  expect_errors=True)

        self.assertEqual(400, resp.status_int)
        self.assertIn("Required param is missing.", resp.json['faultstring'])

    def test_get(self):
        name = 'TestAlias'
        function_version = 0
        body = {'function_id': self.func_id,
                'function_version': function_version,
                'name': name,
                'description': 'new alias'}
        db_api.create_function_alias(**body)

        resp = self.app.get('/v1/aliases/%s' % name)

        context.set_ctx(self.ctx)

        self.assertEqual(200, resp.status_int)
        self.assertEqual("new alias", resp.json.get('description'))

    def test_get_notfound(self):
        resp = self.app.get('/v1/aliases/%s' % 'fake_name',
                            expect_errors=True)

        self.assertEqual(404, resp.status_int)
        self.assertIn("FunctionAlias not found", resp.json['faultstring'])

    def test_get_all(self):
        name = self.rand_name(name="alias", prefix=self.prefix)
        body = {
            'function_id': self.func_id,
            'name': name
        }
        db_api.create_function_alias(**body)

        resp = self.app.get('/v1/aliases')

        self.assertEqual(200, resp.status_int)

        expected = {
            "name": name,
            'function_id': self.func_id,
            'function_version': 0,
            "project_id": unit_base.DEFAULT_PROJECT_ID,
        }
        actual = self._assert_single_item(resp.json['function_aliases'],
                                          name=name)
        self._assertDictContainsSubset(actual, expected)

    def test_delete(self):
        name = self.rand_name(name="alias", prefix=self.prefix)
        function_version = 0
        body = {'function_id': self.func_id,
                'function_version': function_version,
                'name': name,
                'description': 'new alias'}

        db_api.create_function_alias(**body)

        resp = self.app.delete('/v1/aliases/%s' % name)

        self.assertEqual(204, resp.status_int)

        context.set_ctx(self.ctx)

        self.assertRaises(exc.DBEntityNotFoundError,
                          db_api.get_function_alias,
                          name)

    def test_put(self):
        name = self.rand_name(name="alias", prefix=self.prefix)
        function_version = 0
        body = {'function_id': self.func_id,
                'function_version': function_version,
                'name': name,
                'description': 'new alias'}

        db_api.create_function_alias(**body)

        body['function_version'] = 1
        body['description'] = 'update alias'

        resp = self.app.put_json('/v1/aliases/%s' % name, body)

        self.assertEqual(200, resp.status_int)
        self._assertDictContainsSubset(resp.json, body)

    def test_put_without_optional_params(self):
        name = self.rand_name(name="alias", prefix=self.prefix)
        function_version = 1
        body = {'function_id': self.func_id,
                'function_version': function_version,
                'name': name,
                'description': 'new alias'}

        db_api.create_function_alias(**body)

        update_body = {}

        resp = self.app.put_json('/v1/aliases/%s' % name, update_body)

        self.assertEqual(200, resp.status_int)
        self._assertDictContainsSubset(resp.json, body)



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import shutil
import tempfile

import mock
from oslo_config import cfg
import pecan
import pecan.testing
from webtest import app as webtest_app

from qinling.tests.unit import base

CONF = cfg.CONF


class APITest(base.DbTestCase):
    def setUp(self):
        super(APITest, self).setUp()

        # Config package directory before app starts.
        package_dir = tempfile.mkdtemp(prefix='tmp_qinling')
        self.override_config('file_system_dir', package_dir, 'storage')
        self.addCleanup(shutil.rmtree, package_dir, True)

        # Disable authentication by default for API tests.
        self.override_config('auth_enable', False, group='pecan')

        # Disable job handler. The following pecan app instantiation will
        # invoke qinling.api.app:setup_app()
        self.override_config('enable_job_handler', False, group='api')

        pecan_opts = CONF.pecan
        self.app = pecan.testing.load_test_app({
            'app': {
                'root': pecan_opts.root,
                'modules': pecan_opts.modules,
                'debug': pecan_opts.debug,
                'auth_enable': False,
            }
        })

        self.addCleanup(pecan.set_config, {}, overwrite=True)

        self.patch_ctx = mock.patch('qinling.context.Context.from_environ')
        self.mock_ctx = self.patch_ctx.start()
        self.mock_ctx.return_value = self.ctx
        self.addCleanup(self.patch_ctx.stop)

    def _assertNotFound(self, url):
        try:
            self.app.get(url, headers={'Accept': 'application/json'})
        except webtest_app.AppError as error:
            self.assertIn('Bad response: 404 Not Found', str(error))
            return

        self.fail('Expected 404 Not found but got OK')

    def _assertUnauthorized(self, url):
        try:
            self.app.get(url, headers={'Accept': 'application/json'})
        except webtest_app.AppError as error:
            self.assertIn('Bad response: 401 Unauthorized', str(error))
            return

        self.fail('Expected 401 Unauthorized but got OK')



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_config import cfg
from oslo_log import log as logging
import oslo_middleware.cors as cors_middleware
import oslo_middleware.http_proxy_to_wsgi as http_proxy_to_wsgi_middleware
import pecan

from qinling.api import access_control
from qinling import config as q_config
from qinling import context as ctx
from qinling.db import api as db_api
from qinling.services import periodics

LOG = logging.getLogger(__name__)


def get_pecan_config():
    # Set up the pecan configuration.
    opts = cfg.CONF.pecan

    cfg_dict = {
        "app": {
            "root": opts.root,
            "modules": opts.modules,
            "debug": opts.debug,
            "auth_enable": opts.auth_enable
        }
    }

    return pecan.configuration.conf_from_dict(cfg_dict)


def setup_app(config=None):
    if not config:
        config = get_pecan_config()

    q_config.set_config_defaults()

    app_conf = dict(config.app)

    db_api.setup_db()

    if cfg.CONF.api.enable_job_handler:
        LOG.info('Starting periodic tasks...')
        periodics.start_job_handler()

    app = pecan.make_app(
        app_conf.pop('root'),
        hooks=lambda: [ctx.ContextHook(), ctx.AuthHook()],
        logging=getattr(config, 'logging', {}),
        **app_conf
    )

    # Set up access control.
    app = access_control.setup(app)

    # Create HTTPProxyToWSGI wrapper
    app = http_proxy_to_wsgi_middleware.HTTPProxyToWSGI(app, cfg.CONF)

    # Create a CORS wrapper, and attach mistral-specific defaults that must be
    # included in all CORS responses.
    return cors_middleware.CORS(app, cfg.CONF)


def init_wsgi():
    # By default, oslo.config parses the CLI args if no args is provided.
    # As a result, invoking this wsgi script from gunicorn leads to the error
    # with argparse complaining that the CLI options have already been parsed.
    q_config.parse_args(args=[])

    return setup_app()



# Copyright 2018 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import hashlib

from oslo_config import cfg

from qinling import config as qinling_config
from qinling.utils import etcd_util

QINLING_CONF = None


def md5(file=None, content=None):
    hash_md5 = hashlib.md5()

    if file:
        with open(file, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
    elif content:
        hash_md5.update(content)

    return hash_md5.hexdigest()


def get_etcd_client():
    """Use qinling's default CONF to connect to etcd."""
    global QINLING_CONF

    if not QINLING_CONF:
        QINLING_CONF = cfg.ConfigOpts()
        QINLING_CONF(args=[], project='qinling')
        QINLING_CONF.register_opts(qinling_config.etcd_opts,
                                   qinling_config.ETCD_GROUP)

    return etcd_util.get_client(conf=QINLING_CONF)



# Copyright 2018 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from kubernetes.client import api_client
# from kubernetes.client.apis import apps_v1_api
from kubernetes.client.apis import core_v1_api
from kubernetes.client.apis import extensions_v1beta1_api
from kubernetes.client import configuration as k8s_config


def get_k8s_clients(conf):
    config = k8s_config.Configuration()
    config.host = conf.kubernetes.kube_host
    if conf.kubernetes.use_api_certificate:
        config.ssl_ca_cert = conf.kubernetes.ssl_ca_cert
        config.cert_file = conf.kubernetes.cert_file
        config.key_file = conf.kubernetes.key_file
    else:
        config.verify_ssl = False
    client = api_client.ApiClient(configuration=config)
    v1 = core_v1_api.CoreV1Api(client)
    v1extension = extensions_v1beta1_api.ExtensionsV1beta1Api(client)
    # apps_v1 = apps_v1_api.AppsV1Api(client)

    clients = {
        'v1': v1,
        # 'apps_v1': apps_v1
        'v1extension': v1extension
    }

    return clients



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import re

from oslo_config import cfg
from oslo_context import context as oslo_context
import pecan
from pecan import hooks

from qinling import exceptions as exc
from qinling.utils import thread_local

CONF = cfg.CONF
ALLOWED_WITHOUT_AUTH = ['/', '/v1/']
WEBHOOK_REG = '^/v1/webhooks/[a-f0-9-]+/invoke$'
CTX_THREAD_LOCAL_NAME = "QINLING_APP_CTX_THREAD_LOCAL"
DEFAULT_PROJECT_ID = "default"


def authenticate(req):
    # Refer to:
    # https://docs.openstack.org/keystonemiddleware/latest/middlewarearchitecture.html#exchanging-user-information
    identity_status = req.headers.get('X-Identity-Status')
    service_identity_status = req.headers.get('X-Service-Identity-Status')

    if (identity_status == 'Confirmed' or
            service_identity_status == 'Confirmed'):
        return

    if req.headers.get('X-Auth-Token'):
        msg = 'Auth token is invalid: %s' % req.headers['X-Auth-Token']
    else:
        msg = 'Authentication required'

    raise exc.UnauthorizedException(msg)


class AuthHook(hooks.PecanHook):
    def before(self, state):
        if not CONF.pecan.auth_enable:
            return
        if state.request.path in ALLOWED_WITHOUT_AUTH:
            return
        if re.search(WEBHOOK_REG, state.request.path):
            return

        try:
            authenticate(state.request)
        except Exception as e:
            msg = "Failed to validate access token: %s" % str(e)

            pecan.abort(
                status_code=401,
                detail=msg,
                headers={'Server-Error-Message': msg}
            )


def has_ctx():
    return thread_local.has_thread_local(CTX_THREAD_LOCAL_NAME)


def get_ctx():
    if not has_ctx():
        raise exc.ApplicationContextNotFoundException()

    return thread_local.get_thread_local(CTX_THREAD_LOCAL_NAME)


def set_ctx(new_ctx):
    thread_local.set_thread_local(CTX_THREAD_LOCAL_NAME, new_ctx)


class Context(oslo_context.RequestContext):
    def __init__(self, is_trust_scoped=False, trust_id=None, is_admin=False,
                 **kwargs):
        self.is_trust_scoped = is_trust_scoped
        self.trust_id = trust_id

        super(Context, self).__init__(is_admin=is_admin, **kwargs)

    @property
    def projectid(self):
        if CONF.pecan.auth_enable:
            return self.project_id
        else:
            return DEFAULT_PROJECT_ID

    def convert_to_dict(self):
        """Return a dictionary of context attributes.

        Use get_logging_values() instead of to_dict() from parent class to get
        more information from the context. This method is not named "to_dict"
        to avoid recursive calling.
        """
        ctx_dict = self.get_logging_values()
        ctx_dict.update(
            {
                'is_trust_scoped': self.is_trust_scoped,
                'trust_id': self.trust_id,
                'auth_token': self.auth_token,
            }
        )

        return ctx_dict

    @classmethod
    def from_dict(cls, values, **kwargs):
        """Construct a context object from a provided dictionary."""
        kwargs.setdefault(
            'is_trust_scoped', values.get('is_trust_scoped', False)
        )
        kwargs.setdefault('trust_id', values.get('trust_id'))

        return super(Context, cls).from_dict(values, **kwargs)

    @classmethod
    def from_environ(cls, env):
        context = super(Context, cls).from_environ(env)
        context.is_admin = True if 'admin' in context.roles else False

        return context


class ContextHook(hooks.PecanHook):
    def before(self, state):
        context_obj = Context.from_environ(state.request.environ)
        set_ctx(context_obj)

    def after(self, state):
        set_ctx(None)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests http_proxy_to_wsgi middleware."""

from oslo_config import cfg
from oslo_middleware import http_proxy_to_wsgi as http_proxy_to_wsgi_middleware
from qinling.tests.unit.api import base


class TestHTTPProxyToWSGIMiddleware(base.APITest):
    """Test oslo_middleware HTTPProxyToWSGI.

    It checks that oslo_middleware middleware HTTPProxyToWSGI is executed
    when enabled.
    """

    def setUp(self):
        # Make sure the HTTPProxyToWSGI options are registered
        cfg.CONF.register_opts(http_proxy_to_wsgi_middleware.OPTS,
                               'oslo_middleware')

        # Enable proxy headers parsing in HTTPProxyToWSGI middleware.
        self.override_config(
            "enable_proxy_headers_parsing",
            "True",
            group='oslo_middleware'
        )

        # Create the application.
        super(TestHTTPProxyToWSGIMiddleware, self).setUp()



# Copyright 2018 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import collections

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import strutils
import pecan
from pecan import rest
import tenacity
from webob.static import FileIter
import wsmeext.pecan as wsme_pecan

from qinling.api import access_control as acl
from qinling.api.controllers.v1 import resources
from qinling.api.controllers.v1 import types
from qinling import context
from qinling.db import api as db_api
from qinling import exceptions as exc
from qinling import rpc
from qinling.storage import base as storage_base
from qinling.utils import constants
from qinling.utils import etcd_util
from qinling.utils import rest_utils

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class FunctionVersionsController(rest.RestController):
    _custom_actions = {
        'scale_up': ['POST'],
        'scale_down': ['POST'],
        'detach': ['POST'],
    }

    def __init__(self, *args, **kwargs):
        self.type = 'function_version'
        self.storage_provider = storage_base.load_storage_provider(CONF)
        self.engine_client = rpc.get_engine_client()

        super(FunctionVersionsController, self).__init__(*args, **kwargs)

    @tenacity.retry(
        wait=tenacity.wait_fixed(1),
        stop=tenacity.stop_after_attempt(30),
        reraise=True,
        retry=tenacity.retry_if_exception_type(exc.EtcdLockException)
    )
    def _create_function_version(self, project_id, function_id, **kwargs):
        with etcd_util.get_function_version_lock(function_id) as lock:
            if not lock.is_acquired():
                raise exc.EtcdLockException(
                    "Etcd: failed to acquire version lock for function %s." %
                    function_id
                )

            with db_api.transaction():
                # Get latest function package md5 and version number
                func_db = db_api.get_function(function_id, insecure=False)
                if func_db.code['source'] != constants.PACKAGE_FUNCTION:
                    raise exc.NotAllowedException(
                        "Function versioning only allowed for %s type "
                        "function." %
                        constants.PACKAGE_FUNCTION
                    )

                l_md5 = func_db.code['md5sum']
                l_version = func_db.latest_version

                if len(func_db.versions) >= constants.MAX_VERSION_NUMBER:
                    raise exc.NotAllowedException(
                        'Can not exceed maximum number(%s) of versions' %
                        constants.MAX_VERSION_NUMBER
                    )

                # Check if the latest package changed since last version
                changed = self.storage_provider.changed_since(project_id,
                                                              function_id,
                                                              l_md5,
                                                              l_version)
                if not changed:
                    raise exc.NotAllowedException(
                        'Function package not changed since the latest '
                        'version %s.' % l_version
                    )

                LOG.info("Creating %s, function_id: %s, old_version: %d",
                         self.type, function_id, l_version)

                # Create new version and copy package.
                self.storage_provider.copy(project_id, function_id, l_md5,
                                           l_version)
                version = db_api.increase_function_version(function_id,
                                                           l_version,
                                                           **kwargs)
                func_db.latest_version = l_version + 1

            LOG.info("New version %d for function %s created.", l_version + 1,
                     function_id)
            return version

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(
        resources.FunctionVersion,
        types.uuid,
        body=resources.FunctionVersion,
        status_code=201
    )
    def post(self, function_id, body):
        """Create a new version for the function.

        Only allow to create version for package type function.

        The supported boy params:
            - description: Optional. The description of the new version.
        """
        ctx = context.get_ctx()
        acl.enforce('function_version:create', ctx)

        params = body.to_dict()
        values = {
            'description': params.get('description'),
        }

        # Try to create a new function version within lock and db transaction
        try:
            version = self._create_function_version(
                ctx.project_id, function_id, **values
            )
        except exc.EtcdLockException as e:
            LOG.exception(str(e))
            # Reraise a generic exception as the end users should not know
            # the underlying details.
            raise exc.QinlingException('Internal server error.')

        return resources.FunctionVersion.from_db_obj(version)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.FunctionVersions, types.uuid)
    def get_all(self, function_id):
        """Get all the versions of the given function.

        Admin user can get all versions for the normal user's function.
        """
        acl.enforce('function_version:get_all', context.get_ctx())
        LOG.info("Getting versions for function %s.", function_id)

        # Getting function and versions needs to happen in a db transaction
        with db_api.transaction():
            func_db = db_api.get_function(function_id)
            db_versions = func_db.versions

        versions = [resources.FunctionVersion.from_db_obj(v)
                    for v in db_versions]

        return resources.FunctionVersions(function_versions=versions)

    @rest_utils.wrap_pecan_controller_exception
    @pecan.expose()
    @pecan.expose('json')
    def get(self, function_id, version):
        """Get function version or download function version package.

        This method can support HTTP request using either
        'Accept:application/json' or no 'Accept' header.
        """
        ctx = context.get_ctx()
        acl.enforce('function_version:get', ctx)

        download = strutils.bool_from_string(
            pecan.request.GET.get('download', False)
        )
        version = int(version)

        version_db = db_api.get_function_version(function_id, version)

        if not download:
            LOG.info("Getting version %s for function %s.", version,
                     function_id)
            pecan.override_template('json')
            return resources.FunctionVersion.from_db_obj(version_db).to_dict()

        LOG.info("Downloading version %s for function %s.", version,
                 function_id)

        f = self.storage_provider.retrieve(version_db.project_id, function_id,
                                           None, version=version)

        if isinstance(f, collections.Iterable):
            pecan.response.app_iter = f
        else:
            pecan.response.app_iter = FileIter(f)
        pecan.response.headers['Content-Type'] = 'application/zip'
        pecan.response.headers['Content-Disposition'] = (
            'attachment; filename="%s_%s"' % (function_id, version)
        )

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(None, types.uuid, int, status_code=204)
    def delete(self, function_id, version):
        """Delete a specific function version.

        - The version should not being used by any job
        - The version should not being used by any webhook
        - Admin user can not delete normal user's version
        """
        ctx = context.get_ctx()
        acl.enforce('function_version:delete', ctx)
        LOG.info("Deleting version %s of function %s.", version, function_id)

        with db_api.transaction():
            version_db = db_api.get_function_version(function_id, version,
                                                     insecure=False)
            latest_version = version_db.function.latest_version

            version_jobs = db_api.get_jobs(
                function_id=version_db.function_id,
                function_version=version_db.version_number,
                status={'nin': ['done', 'cancelled']}
            )
            if len(version_jobs) > 0:
                raise exc.NotAllowedException(
                    'The function version is still associated with running '
                    'job(s).'
                )

            version_webhook = db_api.get_webhooks(
                function_id=version_db.function_id,
                function_version=version_db.version_number,
            )
            if len(version_webhook) > 0:
                raise exc.NotAllowedException(
                    'The function version is still associated with webhook.'
                )

            filters = rest_utils.get_filters(
                function_id=version_db.function_id,
                function_version=version_db.version_number
            )
            version_aliases = db_api.get_function_aliases(**filters)
            if len(version_aliases) > 0:
                raise exc.NotAllowedException(
                    'The function version is still associated with alias.'
                )

            # Delete resources for function version
            self.engine_client.delete_function(function_id, version=version)
            etcd_util.delete_function(function_id, version=version)

            self.storage_provider.delete(ctx.projectid, function_id, None,
                                         version=version)

            db_api.delete_function_version(function_id, version)

            if latest_version == version:
                version_db.function.latest_version = latest_version - 1

        LOG.info("Version %s of function %s deleted.", version, function_id)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(
        None,
        types.uuid,
        int,
        body=resources.ScaleInfo,
        status_code=202
    )
    def scale_up(self, function_id, version, scale):
        """Scale up the workers for function version execution.

        This is admin only operation. The load monitoring of execution
        depends on the monitoring solution of underlying orchestrator.
        """
        acl.enforce('function_version:scale_up', context.get_ctx())

        func_db = db_api.get_function(function_id)

        # If version=0, it's equivalent to /functions/<funcion-id>/scale_up
        if version > 0:
            db_api.get_function_version(function_id, version)

        params = scale.to_dict()

        LOG.info('Starting to scale up function %s(version %s), params: %s',
                 function_id, version, params)

        self.engine_client.scaleup_function(
            function_id,
            runtime_id=func_db.runtime_id,
            version=version,
            count=params['count']
        )

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(
        None,
        types.uuid,
        int,
        body=resources.ScaleInfo,
        status_code=202
    )
    def scale_down(self, function_id, version, scale):
        """Scale down the workers for function version execution.

        This is admin only operation. The load monitoring of execution
        depends on the monitoring solution of underlying orchestrator.
        """
        acl.enforce('function_version:scale_down', context.get_ctx())

        db_api.get_function(function_id)
        params = scale.to_dict()

        # If version=0, it's equivalent to /functions/<funcion-id>/scale_down
        if version > 0:
            db_api.get_function_version(function_id, version)

        workers = etcd_util.get_workers(function_id, version=version)
        if len(workers) <= 1:
            LOG.info('No need to scale down function %s(version %s)',
                     function_id, version)
            return

        LOG.info('Starting to scale down function %s(version %s), params: %s',
                 function_id, version, params)
        self.engine_client.scaledown_function(function_id, version=version,
                                              count=params['count'])

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(None, types.uuid, int, status_code=202)
    def detach(self, function_id, version):
        """Detach the function version from its underlying workers.

        This is admin only operation, which gives admin user a safe way to
        clean up the underlying resources allocated for the function version.
        """
        acl.enforce('function_version:detach', context.get_ctx())

        db_api.get_function(function_id)
        # If version=0, it's equivalent to /functions/<funcion-id>/detach
        if version > 0:
            db_api.get_function_version(function_id, version)

        LOG.info('Starting to detach function %s(version %s)', function_id,
                 version)

        # Delete allocated resources in orchestrator and etcd keys.
        self.engine_client.delete_function(function_id, version=version)
        etcd_util.delete_function(function_id, version=version)
